#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file DroneFlightDirector.py
 @brief Controlling the flying of drones.
 @date $Date$

 @author East Japan Accounting Center Co., Ltd.

 (c)2017 The University of Aizu
 This software is released under the New BSD License.

"""
import sys
import math
import heapq
import csv
from datetime import datetime

# Import RTM module
import RTC
import OpenRTM_aist

import target_gps
import drone_distance_cal

import DroneServicePort_idl

# Setup logger
from logging import getLogger, StreamHandler
from logging import DEBUG as LOG_LV

sys.path.append(".")

logger = getLogger(__name__)
handler = StreamHandler()
logger.setLevel(LOG_LV)
handler.setLevel(LOG_LV)
logger.addHandler(handler)

# constants
BUTTON_NUM = 14
BUTTON_01_MASK = 0x0001
BUTTON_02_MASK = 0x0002
BUTTON_03_MASK = 0x0004
BUTTON_04_MASK = 0x0008
BUTTON_05_MASK = 0x0010
BUTTON_06_MASK = 0x0020
BUTTON_07_MASK = 0x0040
BUTTON_08_MASK = 0x0080
BUTTON_09_MASK = 0x0100
BUTTON_10_MASK = 0x0200
BUTTON_11_MASK = 0x0400
BUTTON_12_MASK = 0x0800
BUTTON_13_MASK = 0x1000
BUTTON_14_MASK = 0x2000

ANALOG_DIR_UP_DOWN = 0
ANALOG_DIR_LIGHT_LEFT = 1
ANALOG_LSTICK_UP_DOWN = 2
ANALOG_LSTICK_LIGHT_LEFT = 3
ANALOG_RSTICK_UP_DOWN = 4
ANALOG_RSTICK_LIGHT_LEFT = 5
ANALOG_L2 = 6
ANALOG_R2 = 7

INPORT_NUM = 3
SV_RTLALT_NUM = 3
MOVE_SPEED_MAX = 10.0
MOVE_SPEED_MIN = 1.0

MOVE_UPDATE_COEF_ABSOLUTE_NS = 1.0
MOVE_UPDATE_COEF_ABSOLUTE_EW = 0.65
MOVE_UPDATE_COEF_RELATIVE = 0.55

ALLOWABLE_ERROR_POS_CMP = 5.0
ALLOWABLE_ERROR_HEIGHT = 0.0

LEADER_ALTITUDE_CHANGE = 0.2

NO_MOVE_MODE = 0
TRANS_MODE = 1
TURN_MODE = 2

TURN_UPDATE_DIST = 3.0
ONLY_ONE_TURN_SPEED = 45
DRONE_SPEED_RANGE_TO_START_TURN = 1.0

IN_HOME_AREA = 2.0

MODE_ABSOLUTE = 1

ALT_ADJUST_TIME = 3.0

# file mode
FILE_MODE_WRITE = "w"
FILE_MODE_ADD_WRITE = "a"

# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
import Service, Service__POA


# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
droneflightdirector_spec = ["implementation_id", "DroneFlightDirector",
                            "type_name",         "DroneFlightDirector",
                            "description",       "Controlling the flying of drones.",
                            "version",           "1.4.1",
                            "vendor",            "EAC",
                            "category",          "Drone",
                            "activity_type",     "STATIC",
                            "max_instance",      "1",
                            "language",          "Python",
                            "lang_type",         "SCRIPT",
                            "conf.default.AltCoefficient", "1",
                            "conf.default.LatLonCoefficient", "10",
                            "conf.default.LatDiff10m", "0.000137",
                            "conf.default.LonDiff10m", "0.000108",
                            "conf.default.NoseLatDiff10m", "9.01323",
                            "conf.default.NoseLonDiff10m", "8.9831566",
                            "conf.default.MoveSpeed", "3.0",
                            "conf.default.MinAlt", "2.0",
                            "conf.default.MaxAlt", "15.0",
                            "conf.default.DroneNum", "3",
                            "conf.default.Drone1TakeoffAlt", "5.0",
                            "conf.default.Drone2TakeoffAlt", "5.0",
                            "conf.default.Drone3TakeoffAlt", "5.0",
                            "conf.__widget__.AltCoefficient", "spin",
                            "conf.__widget__.LatLonCoefficient", "spin",
                            "conf.__widget__.LatDiff10m", "spin",
                            "conf.__widget__.LonDiff10m", "spin",
                            "conf.__widget__.NoseLatDiff10m", "spin",
                            "conf.__widget__.NoseLonDiff10m", "spin",
                            "conf.__widget__.MoveSpeed", "spin",
                            "conf.__widget__.MinAlt", "spin",
                            "conf.__widget__.MaxAlt", "spin",
                            "conf.__widget__.DroneNum", "spin",
                            "conf.__widget__.Drone1TakeoffAlt", "spin",
                            "conf.__widget__.Drone2TakeoffAlt", "spin",
                            "conf.__widget__.Drone3TakeoffAlt", "spin",
                            "conf.__constraints__.AltCoefficient", "1<=x<=10",
                            "conf.__constraints__.LatLonCoefficient", "2<=x<=10",
                            "conf.__constraints__.LatDiff10m", "0.000000<x<0.100000",
                            "conf.__constraints__.LonDiff10m", "0.000000<x<0.100000",
                            "conf.__constraints__.NoseLatDiff10m", "0.000000<=x<=10.000000",
                            "conf.__constraints__.NoseLonDiff10m", "0.0000000<=x<=10.0000000",
                            "conf.__constraints__.MoveSpeed", "1.0<=x<=10.0",
                            "conf.__constraints__.MinAlt", "1.0<=x<=100.0",
                            "conf.__constraints__.MaxAlt", "10.0<=x<=100.0",
                            "conf.__constraints__.DroneNum", "1<=x<=3",
                            "conf.__constraints__.Drone1TakeoffAlt", "x>=1.0",
                            "conf.__constraints__.Drone2TakeoffAlt", "x>=1.0",
                            "conf.__constraints__.Drone3TakeoffAlt", "x>=1.0",
                            ""]
# </rtc-template>

##
# @class DroneFlightDirector
# @brief Controlling the flying of drones.
#
# Refer to the "機能仕様書_DroneFlightDirector" for details.
#
# Refer to the "機能仕様書_DroneFlightDirector" for details.
#
# Refer to the "機能仕様書_DroneFlightDirector" for details.
#
# Refer to the "機能仕様書_DroneFlightDirector" for details.
#
#


class DroneFlightDirector(OpenRTM_aist.DataFlowComponentBase):

    ##
    # @brief constructor
    # @param manager Maneger Object
    #
    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_ButtonIn = RTC.TimedULong(RTC.Time(0, 0), 0)
        """
              PS4コントローラのボタン入力状態
               - Type: RTC::TimedULong
              """
        self._ButtonInIn = OpenRTM_aist.InPort("ButtonIn", self._d_ButtonIn)
        self._d_AnalogIn = RTC.TimedDoubleSeq(RTC.Time(0, 0), [])
        """
              PS4コントローラのアナログ入力状態
               - Type: RTC::TimedDoubleSeq
              """
        self._AnalogInIn = OpenRTM_aist.InPort("AnalogIn", self._d_AnalogIn)
        self._d_inDrCtrl1 = RTC.DroneCtrl(RTC.Time(0, 0),
                                          0,
                                          RTC.TimedString(RTC.Time(0,0), ""),
                                          RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                      0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                      0.0, 0, RTC.GPS_FIX_NORMAL),
                                          RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                      0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                      0.0, 0, RTC.GPS_FIX_NORMAL),
                                          RTC.TimedFloatSeq(RTC.Time(0, 0), [0.0, 0.0, 0.0]),
                                          RTC.TimedFloat(RTC.Time(0, 0), 0),
                                          RTC.TimedString(RTC.Time(0, 0), ""))
        """
              ドローン1号機のコントロール情報。以下の情報をパッケージ。
              ・受信ID(DroneIOからの送信ID)
              ・ホーム(基地局)座標
              ・ドローン現在位置座標
              ・加速度
              ・磁気コンパス
              ・ドローンI/Oの内部ステータス
               - Type: RTC::DroneCtrl
              """
        self._inDrCtrl1In = OpenRTM_aist.InPort("inDrCtrl1", self._d_inDrCtrl1)
        self._d_inDrCtrl2 = RTC.DroneCtrl(RTC.Time(0, 0),
                                          0,
                                          RTC.TimedString(RTC.Time(0,0), ""),
                                          RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                      0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                      0.0, 0, RTC.GPS_FIX_NORMAL),
                                          RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                      0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                      0.0, 0, RTC.GPS_FIX_NORMAL),
                                          RTC.TimedFloatSeq(RTC.Time(0, 0), [0.0, 0.0, 0.0]),
                                          RTC.TimedFloat(RTC.Time(0, 0), 0),
                                          RTC.TimedString(RTC.Time(0, 0), ""))
        """
              ドローン2号機のコントロール情報。以下の情報をパッケージ。
              ・受信ID(DroneIOからの送信ID)
              ・ホーム(基地局)座標
              ・ドローン現在位置座標
              ・加速度
              ・磁気コンパス
              ・ドローンI/Oの内部ステータス
               - Type: RTC::DroneCtrl
              """
        self._inDrCtrl2In = OpenRTM_aist.InPort("inDrCtrl2", self._d_inDrCtrl2)
        self._d_inDrCtrl3 = RTC.DroneCtrl(RTC.Time(0, 0),
                                          0,
                                          RTC.TimedString(RTC.Time(0,0), ""),
                                          RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                      0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                      0.0, 0, RTC.GPS_FIX_NORMAL),
                                          RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                      0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                      0.0, 0, RTC.GPS_FIX_NORMAL),
                                          RTC.TimedFloatSeq(RTC.Time(0, 0), [0.0, 0.0, 0.0]),
                                          RTC.TimedFloat(RTC.Time(0, 0), 0),
                                          RTC.TimedString(RTC.Time(0, 0), ""))
        """
              ドローン3号機のコントロール情報。以下の情報をパッケージ。
              ・受信ID(DroneIOからの送信ID)
              ・ホーム(基地局)座標
              ・ドローン現在位置座標
              ・加速度
              ・磁気コンパス
              ・ドローンI/Oの内部ステータス
               - Type: RTC::DroneCtrl
              """
        self._inDrCtrl3In = OpenRTM_aist.InPort("inDrCtrl3", self._d_inDrCtrl3)
        self._d_outVirtualLeader = RTC.DroneMap(RTC.Time(0, 0),
                                                RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                            91.0, 181.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                            0.0, 0, RTC.GPS_FIX_NORMAL),
                                                RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                            91.0, 181.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                            0.0, 0, RTC.GPS_FIX_NORMAL),
                                                RTC.TimedFloat(RTC.Time(0, 0), 0),
                                                RTC.TimedString(RTC.Time(0, 0), "State"))
        """
              マップ上に仮想リーダー機の現在地を表示するための情報。以下の情報をパッケージ
              。
              ・ドローン現在地座標
              ・ターゲット座標
              ・磁気コンパス
              ・ドローンI/Oの内部ステータス
               - Type: RTC::DroneMap
              """
        self._outVirtualLeaderOut = OpenRTM_aist.OutPort("outVirtualLeader", self._d_outVirtualLeader)
        self._d_outTarget1 = RTC.DroneTarget(RTC.Time(0, 0),
                                             0,
                                             RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                         0.0, 0, RTC.GPS_FIX_NORMAL),
                                             RTC.TimedFloat(RTC.Time(0, 0), 0),
                                             RTC.TimedShortSeq(RTC.Time(0, 0), [0, 0]))
        """
              ドローン1号機のターゲット情報。以下の情報をパッケージ。
              ・送信ID(1から始まる連番)
              ・ターゲット座標
              ・ターゲット速度
              ・旋回方向及び移動量
               - Type: RTC::DroneTarget
              """
        self._outTarget1Out = OpenRTM_aist.OutPort("outTarget1", self._d_outTarget1)
        self._d_outTarget2 = RTC.DroneTarget(RTC.Time(0, 0),
                                             0,
                                             RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                         0.0, 0, RTC.GPS_FIX_NORMAL),
                                             RTC.TimedFloat(RTC.Time(0, 0), 0),
                                             RTC.TimedShortSeq(RTC.Time(0, 0), [0, 0]))
        """
              ドローン2号機のターゲット情報。以下の情報をパッケージ。
              ・送信ID(1から始まる連番)
              ・ターゲット座標
              ・ターゲット速度
              ・旋回方向及び移動量
               - Type: RTC::DroneTarget
              """
        self._outTarget2Out = OpenRTM_aist.OutPort("outTarget2", self._d_outTarget2)
        self._d_outTarget3 = RTC.DroneTarget(RTC.Time(0, 0),
                                             0,
                                             RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                         0.0, 0, RTC.GPS_FIX_NORMAL),
                                             RTC.TimedFloat(RTC.Time(0, 0), 0),
                                             RTC.TimedShortSeq(RTC.Time(0, 0), [0, 0]))
        """
              ドローン3号機のターゲット情報。以下の情報をパッケージ。
              ・送信ID(1から始まる連番)
              ・ターゲット座標
              ・ターゲット速度
              ・旋回方向及び移動量
               - Type: RTC::DroneTarget
              """
        self._outTarget3Out = OpenRTM_aist.OutPort("outTarget3", self._d_outTarget3)
        self._d_outSettingSpeed = RTC.TimedFloat(RTC.Time(0, 0), 0.0)
        """
              User setting speed
               - Type: RTC::TimedFloat
               - Unit: m/sec
              """
        self._outSettingSpeedOut = OpenRTM_aist.OutPort("outSettingSpeed", self._d_outSettingSpeed)
        self._d_outMoveMode = RTC.TimedString(RTC.Time(0,0),0)
        """
              移動モード（Absolute or Relative）
               - Type: RTC::TimedString
              """
        self._outMoveModeOut = OpenRTM_aist.OutPort("outMoveMode", self._d_outMoveMode)

        """
        Acquire the drone info of drone1
        """
        self._ReqDroneInfo1Port = OpenRTM_aist.CorbaPort("ReqDroneInfo1")
        """
        Acquire the drone info of drone2
        """
        self._ReqDroneInfo2Port = OpenRTM_aist.CorbaPort("ReqDroneInfo2")
        """
        Acquire the drone info of drone3
        """
        self._ReqDroneInfo3Port = OpenRTM_aist.CorbaPort("ReqDroneInfo3")

        """
        Acquire the RTL altitude of drone1
         - Return Value:  RTL altitude
        """
        self._getRTLAlt1 = OpenRTM_aist.CorbaConsumer(interfaceType=Service.RTLAlt)
        """
        Acquire the RTL altitude of drone2
         - Return Value:  RTL altitude
        """
        self._getRTLAlt2 = OpenRTM_aist.CorbaConsumer(interfaceType=Service.RTLAlt)
        """
        Acquire the RTL altitude of drone3
         - Return Value:  RTL altitude
        """
        self._getRTLAlt3 = OpenRTM_aist.CorbaConsumer(interfaceType=Service.RTLAlt)

        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">
        """
              移動ターゲット高度算出係数
               - Name: AltCoefficient AltCoefficient
               - DefaultValue: 1
               - Range: 1<=x<=10
              """
        self._AltCoefficient = [1]
        """
              移動ターゲット経緯度算出係数
               - Name: LatLonCoefficient LatLonCoefficient
               - DefaultValue: 10
               - Range: 2<=x<=10
              """
        self._LatLonCoefficient = [10]
        """
              10mあたりの緯度差分
               - Name: LatDiff10m LatDiff10m
               - DefaultValue: 0.000137
               - Range: 0.000000<x<0.100000
              """
        self._LatDiff10m = [0.000137]
        """
              10mあたりの経度差分
               - Name: LonDiff10m LonDiff10m
               - DefaultValue: 0.000108
               - Range: 0.000000<x<0.100000
              """
        self._LonDiff10m = [0.000108]
        """
              10mあたりの機首緯度差分
               - Name: NoseLatDiff10m NoseLatDiff10m
               - DefaultValue: 9.01323
               - Range: 0.000000<=x<=10.000000
              """
        self._NoseLatDiff10m = [9.01323]
        """
              10mあたりの機首経度差分
               - Name: NoseLonDiff10m NoseLonDiff10m
               - DefaultValue: 8.9831566
               - Range: 0.0000000<=x<=10.0000000
              """
        self._NoseLonDiff10m = [8.9831566]
        """
              ドローン移動速度
               - Name: MoveSpeed MoveSpeed
               - DefaultValue: 3.0
               - Range: 1.0<=x<=10.0
              """
        self._MoveSpeed = [3.0]
        """
              ドローン最小高度
               - Name: MinAlt MinAlt
               - DefaultValue: 2.0
               - Range: 1.0<=x<=100.0
              """
        self._MinAlt = [2.0]
        """
              ドローン最大高度
              ※ドローン最大高度(MaxAlt)の設定値は、最小高度(MinAlt)の設定値以上の値とする
              。
               - Name: MaxAlt MaxAlt
               - DefaultValue: 15.0
               - Range: 10.0<=x<=100.0
              """
        self._MaxAlt = [15.0]
        """
              制御するドローン台数
               - Name: DroneNum DroneNum
               - DefaultValue: 3
               - Range: 1<=x<=3
              """
        self._DroneNum = [3]
        """
              ドローン離陸高度　（１号機）
               - Name: Drone1TakeoffAlt Drone1TakeoffAlt
               - DefaultValue: 5.0
               - Range: 1.0<=x
              """
        self._Drone1TakeoffAlt = [5.0]
        """
              ドローン離陸高度　（２号機）
               - Name: Drone2TakeoffAlt Drone2TakeoffAlt
               - DefaultValue: 5.0
               - Range: 1.0<=x
              """
        self._Drone2TakeoffAlt = [5.0]
        """
              ドローン離陸高度　（３号機）
               - Name: Drone3TakeoffAlt Drone3TakeoffAlt
               - DefaultValue: 5.0
               - Range: 1.0<=x
              """
        self._Drone3TakeoffAlt = [5.0]

        # </rtc-template>

        self._d_inDrCtrl = [0] * INPORT_NUM
        self._inDrCtrlIn = [0] * INPORT_NUM

        self._d_outTarget = [0] * INPORT_NUM
        self._outTargetOut = [0] * INPORT_NUM

        self._d_inDrCtrl[0] = self._d_inDrCtrl1
        self._inDrCtrlIn[0] = self._inDrCtrl1In
        self._d_inDrCtrl[1] = self._d_inDrCtrl2
        self._inDrCtrlIn[1] = self._inDrCtrl2In
        self._d_inDrCtrl[2] = self._d_inDrCtrl3
        self._inDrCtrlIn[2] = self._inDrCtrl3In

        self._d_outTarget[0] = self._d_outTarget1
        self._outTargetOut[0] = self._outTarget1Out
        self._d_outTarget[1] = self._d_outTarget2
        self._outTargetOut[1] = self._outTarget2Out
        self._d_outTarget[2] = self._d_outTarget3
        self._outTargetOut[2] = self._outTarget3Out

        self._button = False
        self._analog = False
        self._drone_gps = [False] * INPORT_NUM
        self._dr_ctrl_rcv = [False] * INPORT_NUM

        self._target_gps = target_gps.TargetGPS()

        self._leader_set_first = True
        self._rtl_num = 0

        self._drone_diff_pos = [False] * INPORT_NUM
        self._leader_diff_pos = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                            0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                            0.0, 0, RTC.GPS_FIX_NORMAL)
        for num in range(INPORT_NUM):
            self._drone_diff_pos[num] = RTC.GPSData(
                RTC.Time(0,0), RTC.GPSTime(0,0),
                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                0.0, 0, RTC.GPS_FIX_NORMAL)

        self._leader_home = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                       0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                       0.0, 0, RTC.GPS_FIX_NORMAL)

        self._leader_pos = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                       0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                       0.0, 0, RTC.GPS_FIX_NORMAL)
        self._leader_target = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                          0.0, 0, RTC.GPS_FIX_NORMAL)

        self._avg_drone_pos = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                          0.0, 0, RTC.GPS_FIX_NORMAL)

        self._target = [None] * INPORT_NUM
        for num in range(INPORT_NUM):
            self._target[num] = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                            0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                            0.0, 0, RTC.GPS_FIX_NORMAL)

        self._mode = NO_MOVE_MODE

        self._last_front_back = 0
        self._last_right_left = 0
        self._last_up_down = 0
        self._last_speedparam = 0
        self._last_move_mode = "Absolute"

        self._avg_mag = 0
        self._turn_angle = 0
        self._move_mag = 0

        self._Drone_Num = self._DroneNum[0]
        self._SpeedParam = self._MoveSpeed[0]

        self._takeoff_alt = [None] * INPORT_NUM
        self._takeoff_alt[0] = self._Drone1TakeoffAlt[0]
        self._takeoff_alt[1] = self._Drone2TakeoffAlt[0]
        self._takeoff_alt[2] = self._Drone3TakeoffAlt[0]

        self._all_armed = False
        self._pivot = [0] * 2

        self._home = False

        self._alt_adjust_cnt = 0

        self._file_name = ""
        self._file = None
        self._writer = None
        self._header = ["Category","ActStartDate","ActStartTime","RcvDate","RcvTime","RcvRobotID","RcvID",
                       "HomeLat","HomeLon","HomeAlt","RobotLat","RobotLon","RobotAlt",
                       "AccX","AccY","AccZ","Mag","State","SndDate","SndTime","SndRobotID","SndID","MoveMode",
                       "TargetLat","TargetLon","TargetAlt","TargetSpd","TargetYawDir","TargetYawDeg","RTCState"]
        self._act_start_date = ""
        self._act_start_time = ""

        self._out_target_wrote = [False] * INPORT_NUM

        # connection check for data port ButtonInIn
        func = lambda: self._cb_connect_status_change("ButtonIn", True)
        ob = ConnectionCallback(func)
        self._ButtonInIn.setOnConnected(ob)

        func = lambda: self._cb_connect_status_change("ButtonIn", False)
        ob = ConnectionCallback(func)
        self._ButtonInIn.setOnDisconnected(ob)

        # connection check for data port AnalogInIn
        func = lambda: self._cb_connect_status_change("AnalogIn", True)
        ob = ConnectionCallback(func)
        self._AnalogInIn.setOnConnected(ob)

        func = lambda: self._cb_connect_status_change("AnalogIn", False)
        ob = ConnectionCallback(func)
        self._AnalogInIn.setOnDisconnected(ob)

    ##
    # RTC初期化
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry()
    #
    # @return RTC::ReturnCode_t
    #
    #
    def onInitialize(self):
        # Bind variables and configuration variable
        self.bindParameter("AltCoefficient", self._AltCoefficient, "1")
        self.bindParameter("LatLonCoefficient", self._LatLonCoefficient, "10")
        self.bindParameter("LatDiff10m", self._LatDiff10m, "0.000137")
        self.bindParameter("LonDiff10m", self._LonDiff10m, "0.000108")
        self.bindParameter("NoseLatDiff10m", self._NoseLatDiff10m, "9.01323")
        self.bindParameter("NoseLonDiff10m", self._NoseLonDiff10m, "8.9831566")
        self.bindParameter("MoveSpeed", self._MoveSpeed, "3.0")
        self.bindParameter("MinAlt", self._MinAlt, "2.0")
        self.bindParameter("MaxAlt", self._MaxAlt, "15.0")
        self.bindParameter("DroneNum", self._DroneNum, "3")
        self.bindParameter("Drone1TakeoffAlt", self._Drone1TakeoffAlt, "5.0")
        self.bindParameter("Drone2TakeoffAlt", self._Drone2TakeoffAlt, "5.0")
        self.bindParameter("Drone3TakeoffAlt", self._Drone3TakeoffAlt, "5.0")

        # Set InPort buffers
        self.addInPort("ButtonIn", self._ButtonInIn)
        self.addInPort("AnalogIn", self._AnalogInIn)
        self.addInPort("inDrCtrl1", self._inDrCtrl1In)
        self.addInPort("inDrCtrl2", self._inDrCtrl2In)
        self.addInPort("inDrCtrl3", self._inDrCtrl3In)

        # Set OutPort buffers
        self.addOutPort("outVirtualLeader", self._outVirtualLeaderOut)
        self.addOutPort("outTarget1", self._outTarget1Out)
        self.addOutPort("outTarget2", self._outTarget2Out)
        self.addOutPort("outTarget3", self._outTarget3Out)
        self.addOutPort("outSettingSpeed", self._outSettingSpeedOut)
        self.addOutPort("outMoveMode", self._outMoveModeOut)

        # Set service provider to Ports

        # Set service consumers to Ports
        self._ReqDroneInfo1Port.registerConsumer("getRTLAlt1", "Service::RTLAlt", self._getRTLAlt1)
        self._ReqDroneInfo2Port.registerConsumer("getRTLAlt2", "Service::RTLAlt", self._getRTLAlt2)
        self._ReqDroneInfo3Port.registerConsumer("getRTLAlt3", "Service::RTLAlt", self._getRTLAlt3)

        # Set CORBA Service Ports
        self.addPort(self._ReqDroneInfo1Port)
        self.addPort(self._ReqDroneInfo2Port)
        self.addPort(self._ReqDroneInfo3Port)

        return RTC.RTC_OK

        ##
        #
        # The finalize action (on ALIVE->END transition)
        # formaer rtc_exiting_entry()
        #
        # @return RTC::ReturnCode_t

        #
    def onFinalize(self):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()
        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Exit")
        self._close_rtc_viewer_log()

        return RTC.RTC_OK

    #       ##
    #       #
    #       # The startup action when ExecutionContext startup
    #       # former rtc_starting_entry()
    #       #
    #       # @param ec_id target ExecutionContext Id
    #       #
    #       # @return RTC::ReturnCode_t
    #       #
    #       #
    # def onStartup(self, ec_id):
    #
    #       return RTC.RTC_OK

    #       ##
    #       #
    #       # The shutdown action when ExecutionContext stop
    #       # former rtc_stopping_entry()
    #       #
    #       # @param ec_id target ExecutionContext Id
    #       #
    #       # @return RTC::ReturnCode_t
    #       #
    #       #
    # def onShutdown(self, ec_id):
    #
    #       return RTC.RTC_OK

        ##
        # 変数初期化
        #
        # The activated action (Active state entry action)
        # former rtc_active_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onActivated(self, ec_id):

        self._Drone_Num = self._DroneNum[0]
        self._SpeedParam = self._MoveSpeed[0]
        self._leader_set_first = True
        self._rtl_num = 0
        self._last_move_mode = "Absolute"
        self._all_armed = False
        self._pivot = [0] * 2

        self._file_name = ""
        self._file = None
        self._writer = None
        self._act_start_date, self._act_start_time = self._get_datetime_to_str()

        if self._is_landing() == False:
            self._rtl_instruction = False
            self._drone_gps = [False] * INPORT_NUM

            # init leader info
            self._leader_pos = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                        100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                        0.0, 0, RTC.GPS_FIX_NORMAL)
            self._leader_target = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                            100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                            0.0, 0, RTC.GPS_FIX_NORMAL)
            # init drone info
            for num in range(self._Drone_Num):
                self._d_inDrCtrl[num].state.data = "Disarm"
                self._d_outTarget[num].yaw.data = [0,0]
                self._target[num] = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                                100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                0.0, 0, RTC.GPS_FIX_NORMAL)

        self._execute_rate = self.getExecutionRate(ec_id)

        self._alt_adjust_cnt = self._execute_rate * ALT_ADJUST_TIME
        
        init_movemode = MODE_ABSOLUTE
        self._target_gps.set_controller_mode(init_movemode)

        # viewer-log open and write(header-row)
        self._log_row = [["NULL" for item in range(len(self._header))] for robot_num in range(self._Drone_Num)]
        self._ready_to_log_write(FILE_MODE_WRITE)

        if self._takeoff_range_check() == False:
            return RTC.RTC_ERROR

        return RTC.RTC_OK

        ##
        #
        # The deactivated action (Active state exit action)
        # former rtc_active_exit()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onDeactivated(self, ec_id):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()
        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Inactive")
        self._close_rtc_viewer_log()

        return RTC.RTC_OK

        ##
        # PS4キー入力、ドローンステータス取得、およびコントロールデータ送信の周期処理。
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onExecute(self, ec_id):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()

        self._read_inport()
        self._is_drone_on_home()
        self._move_speed()
        self._write_outport()

        self._log_row = [["NULL" for item in range(len(self._header))] for robot_num in range(self._Drone_Num)]
        self._write_rtc_viewer_log("Active")

        return RTC.RTC_OK

    #       ##
    #       #
    #       # The aborting action when main logic error occurred.
    #       # former rtc_aborting_entry()
    #       #
    #       # @param ec_id target ExecutionContext Id
    #       #
    #       # @return RTC::ReturnCode_t
    #       #
    #       #
    # def onAborting(self, ec_id):
    #
    #       return RTC.RTC_OK

        ##
        #
        # The error action in ERROR state
        # former rtc_error_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onError(self, ec_id):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()

        if self._file is None:
            self._ready_to_log_write(FILE_MODE_ADD_WRITE)

        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Error")

        return RTC.RTC_OK

        ##
        #
        # The reset action that is invoked resetting
        # This is same but different the former rtc_init_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onReset(self, ec_id):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()
        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Inactive")
        self._close_rtc_viewer_log()

        return RTC.RTC_OK

    #       ##
    #       #
    #       # The state update action that is invoked after onExecute() action
    #       # no corresponding operation exists in OpenRTm-aist-0.2.0
    #       #
    #       # @param ec_id target ExecutionContext Id
    #       #
    #       # @return RTC::ReturnCode_t
    #       #

    #       #
    # def onStateUpdate(self, ec_id):
    #
    #       return RTC.RTC_OK

    #       ##
    #       #
    #       # The action that is invoked when execution context's rate is changed
    #       # no corresponding operation exists in OpenRTm-aist-0.2.0
    #       #
    #       # @param ec_id target ExecutionContext Id
    #       #
    #       # @return RTC::ReturnCode_t
    #       #
    #       #
    # def onRateChanged(self, ec_id):
    #
    #       return RTC.RTC_OK


    def _read_inport(self):
        
        # button
        self._d_ButtonIn = RTC.TimedULong(RTC.Time(0, 0), 0)

        if self._button:
            self._d_ButtonIn = self._ButtonInIn.read()
            if self._is_valid_button(self._d_ButtonIn.data):
                logger.debug("Button = 0b" + format(self._d_ButtonIn.data, '016b'))
            else:
                self._d_ButtonIn.data = 0
        else:
            logger.error("ERROR: ButtonIn is NOT connected")

        # analog
        self._d_AnalogIn = RTC.TimedDoubleSeq(RTC.Time(0, 0), [])

        if self._analog:
            self._d_AnalogIn = self._AnalogInIn.read()
            for i in range(0, len(self._d_AnalogIn.data), 1):
                if self._is_valid_analog(self._d_AnalogIn.data[i], i):
                    logger.debug("Analog[%d]:  % f", i, self._d_AnalogIn.data[i])
                else:
                    self._d_AnalogIn.data[i] = 0
        else:
            logger.error("ERROR: AnalogIn is NOT connected")

        # drone_ctrl
        for num in range(self._Drone_Num):
            if self._inDrCtrlIn[num].isNew():
                prev_inDrCtrl = self._d_inDrCtrl[num]
                self._d_inDrCtrl[num] = self._inDrCtrlIn[num].read()
                OpenRTM_aist.setTimestamp(self._d_inDrCtrl[num])
                self._dr_ctrl_rcv[num] = True

                if self._is_valid_drctrl(self._d_inDrCtrl[num], num):
                    self._drone_gps[num] = True

                    logger.debug("Rcv ID: % d", self._d_inDrCtrl[num].id)

                    logger.debug("RobotID: %s", self._d_inDrCtrl[num].robotID.data)

                    logger.debug("Home%dGPS:  % f % f % f",
                                 (num + 1),
                                 self._d_inDrCtrl[num].homeGPS.latitude,
                                 self._d_inDrCtrl[num].homeGPS.longitude,
                                 self._d_inDrCtrl[num].homeGPS.altitude)

                    logger.debug("Drone%dGPS:  % f % f % f",
                                 (num + 1),
                                 self._d_inDrCtrl[num].droneGPS.latitude,
                                 self._d_inDrCtrl[num].droneGPS.longitude,
                                 self._d_inDrCtrl[num].droneGPS.altitude)

                    logger.debug("Accel%d:  % f % f % f",
                                 (num + 1),
                                 self._d_inDrCtrl[num].accel.data[0],
                                 self._d_inDrCtrl[num].accel.data[1],
                                 self._d_inDrCtrl[num].accel.data[2])

                    logger.debug("Mag%d:  % d",
                                 (num + 1),
                                 self._d_inDrCtrl[num].mag.data)

                    logger.debug("State%d:  % s",
                                 (num + 1),
                                 self._d_inDrCtrl[num].state.data)
                else:
                    self._d_inDrCtrl[num] = prev_inDrCtrl
            else:
                self._dr_ctrl_rcv[num] = False

    def _set_outport(self):
        self._set_move_mode()
        if self._is_rcv_inport_valid_val():
            self._set_target_gps()
            self._set_target_speed()
            self._set_target_yaw()
        else:
            for num in range(self._Drone_Num):
                if self._drone_gps[num]:
                    self._d_outTarget[num].pos.latitude = self._d_inDrCtrl[num].droneGPS.latitude
                    self._d_outTarget[num].pos.longitude = self._d_inDrCtrl[num].droneGPS.longitude
                    self._d_outTarget[num].pos.altitude = self._d_inDrCtrl[num].droneGPS.altitude
                    self._d_outTarget[num].yaw.data = [0, int(self._d_inDrCtrl[num].mag.data)]

        self._set_leader_target_gps()

    def _write_outport(self):

        self._out_target_wrote = [False] * INPORT_NUM

        # 計算前のself._d_outTargetを保存する
        prev_outTarget = [0] * self._Drone_Num
        for num in range(self._Drone_Num):
            prev_outTarget[num] = self._d_outTarget[num]

        if self._rtl_instruction:
            # write virtual leader
            if self._is_valid_virtual_leader(self._d_outVirtualLeader):
                self._outVirtualLeaderOut.write()

            for num in range(self._Drone_Num):
                if self._home:
                    # set home altitude when all drone arrive over the home position
                    self._d_outTarget[num].pos.altitude = self._d_inDrCtrl[num].homeGPS.altitude
                else:
                    if self._d_inDrCtrl[num].state.data != "RTL":
                        # set home altitude until the drone receives the RTL instruction and the states become RTL
                        self._d_outTarget[num].pos.altitude = self._d_inDrCtrl[num].homeGPS.altitude
                    else:
                        # set RTL altitude until the all drone arrive over the home position when after state RTL
                        self._d_outTarget[num].pos.altitude = self._rtl_alt[num]

                self._d_outTarget[num].id += 1
                OpenRTM_aist.setTimestamp(self._d_outTarget[num])
                if self._outTargetOut[num].write():
                    self._out_target_wrote[num] = True
                    logger.debug("Send ID: % d", self._d_outTarget[num].id)
                    logger.debug("Target%dGPS: % f % f % f",
                                (num + 1),
                                self._d_outTarget[num].pos.latitude,
                                self._d_outTarget[num].pos.longitude,
                                self._d_outTarget[num].pos.altitude)

        else:
            # set
            self._set_outport()

            # write
            # virtual leader
            if self._is_valid_virtual_leader(self._d_outVirtualLeader):
                self._outVirtualLeaderOut.write()
            # drone
            for num in range(self._Drone_Num):
                if self._drone_gps[num]:
                    if self._is_valid_target(self._d_outTarget[num], num) == False:
                        self._d_outTarget[num] = prev_outTarget[num]
                    self._d_outTarget[num].id += 1
                    OpenRTM_aist.setTimestamp(self._d_outTarget[num])
                    if self._outTargetOut[num].write():
                        self._out_target_wrote[num] = True

            self._outSettingSpeedOut.write()
            OpenRTM_aist.setTimestamp(self._d_outMoveMode)
            self._outMoveModeOut.write()

            logger.debug("VirtualLeader Pos: % f % f % f",
                         self._d_outVirtualLeader.droneGPS.latitude,
                         self._d_outVirtualLeader.droneGPS.longitude,
                         self._d_outVirtualLeader.droneGPS.altitude)
            logger.debug("VirtualLeader Target: % f % f % f",
                         self._d_outVirtualLeader.targetGPS.latitude,
                         self._d_outVirtualLeader.targetGPS.longitude,
                         self._d_outVirtualLeader.targetGPS.altitude)
            logger.debug("VirtualLeader Mag: % f", self._d_outVirtualLeader.mag.data),
            logger.debug("VirtualLeader State: % s", self._d_outVirtualLeader.state.data)

            for num in range(self._Drone_Num):
                if self._out_target_wrote[num]:
                    logger.debug("Send ID: % d", self._d_outTarget[num].id)
                    logger.debug("Target%dGPS: % f % f % f",
                                (num + 1),
                                 self._d_outTarget[num].pos.latitude,
                                 self._d_outTarget[num].pos.longitude,
                                 self._d_outTarget[num].pos.altitude)
                    logger.debug("Target%dSpeed: %f",
                                 (num + 1),
                                 self._d_outTarget[num].speed.data)
                    logger.debug("Target%dYaw: (dir)%d, (deg)%d",
                                (num + 1),
                                 self._d_outTarget[num].yaw.data[0],
                                 self._d_outTarget[num].yaw.data[1])
            logger.debug("MoveMode: %s",
                         self._d_outMoveMode.data)

            self._last_move_mode = self._d_outMoveMode.data
            self._last_speedparam = self._SpeedParam          

    def _is_rcv_inport_valid_val(self):
        ret = False

        if self._all_armed == False:
            self._is_all_drone_states_armed()
        else:
            for num in range(self._Drone_Num):
                if self._d_inDrCtrl[num].state.data in ["Armed", "Takeoff", "In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"]:
                    if self._drone_gps[num]:
                        ret = True
                    else:
                        logger.error("ERROR: GPS data is NOT available")
                        ret = False
                        break

        return ret

    def _is_all_drone_states_armed(self):
        armed = False
        for num in range(self._Drone_Num):
            if self._d_inDrCtrl[num].state.data == "Armed":
                armed = True
            else:
                armed = False
                break

        self._all_armed = armed

    def _config_target_gps(self):
        config = {"AltCoefficient": self._AltCoefficient[0],
                  "LatLonCoefficient": self._LatLonCoefficient[0],
                  "LatDiff10m": self._LatDiff10m[0],
                  "LonDiff10m": self._LonDiff10m[0],
                  "NoseLatDiff10m": self._NoseLatDiff10m[0],
                  "NoseLonDiff10m": self._NoseLonDiff10m[0],
                  "MoveSpeed": self._MoveSpeed[0],
                  "MinAlt": self._MinAlt[0] - min(self._alt_diff),
                  "MaxAlt": self._MaxAlt[0] - max(self._alt_diff),
                  "DroneNum": self._Drone_Num
                  }

        self._target_gps.config_param(config)

    def _set_leader(self):
        count = 0
        for num in range(self._Drone_Num):
            if self._d_inDrCtrl[num].state.data in ["RTL", "Disarm"]:
                count += 1

        if self._rtl_num < count:
            self._rtl_num = count

        self._calc_leader_pos()

        if self._leader_set_first:
            self._leader_set_first = False
            self._set_leader_pos()
            self._drone_diff()
            # pivot機算出
            self._select_pivot()
            self._mode = NO_MOVE_MODE

    def _calc_leader_pos(self):
        self._avg_drone_pos.latitude = 0.0
        self._avg_drone_pos.longitude = 0.0
        self._avg_drone_pos.altitude = 0.0
        self._avg_mag = 0
        self._avg_speed = 0

        mag_normal_result = None

        for num in range(self._Drone_Num):
            if self._d_inDrCtrl[num].state.data in ["Armed", "Takeoff", "In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"]:
                # sum of drone_pos
                self._avg_drone_pos.latitude += self._d_inDrCtrl[num].droneGPS.latitude
                self._avg_drone_pos.longitude += self._d_inDrCtrl[num].droneGPS.longitude
                if self._d_inDrCtrl[num].droneGPS.altitude < self._MinAlt[0]:
                    self._avg_drone_pos.altitude += self._takeoff_alt[num]
                else:
                    self._avg_drone_pos.altitude += self._d_inDrCtrl[num].droneGPS.altitude
                # sum of mag
                if mag_normal_result == None:
                    mag_normal_result = self._d_inDrCtrl[num].mag.data
                else:
                    diff = self._d_inDrCtrl[num].mag.data - self._d_inDrCtrl[last_calc_drone_num].mag.data
                    if diff < -180:
                        diff += 360
                    elif diff > 180:
                        diff -= 360
                    mag_normal_result += diff
                self._avg_mag += mag_normal_result
                last_calc_drone_num = num

                # sum of speed
                self._avg_speed += math.sqrt((self._d_inDrCtrl[num].accel.data[0] * self._d_inDrCtrl[num].accel.data[0]) +
                                             (self._d_inDrCtrl[num].accel.data[1] * self._d_inDrCtrl[num].accel.data[1]))

        # average of drone_pos
        self._avg_drone_pos.latitude /= (self._Drone_Num - self._rtl_num)
        self._avg_drone_pos.longitude /= (self._Drone_Num - self._rtl_num)
        self._avg_drone_pos.altitude /= (self._Drone_Num - self._rtl_num)

        # average of mag
        self._avg_mag /= (self._Drone_Num - self._rtl_num)
        if self._avg_mag > 360:
            self._avg_mag = self._avg_mag % 360
        elif self._avg_mag < 0:
            self._avg_mag = 360 - (abs(self._avg_mag) % 360)

        # average of speed
        self._avg_speed /= (self._Drone_Num - self._rtl_num)

    def _set_leader_pos(self):
        # set leader home
        self._leader_home.latitude = self._avg_drone_pos.latitude
        self._leader_home.longitude = self._avg_drone_pos.longitude
        self._leader_home.altitude = self._avg_drone_pos.altitude

        # set leader pos
        self._leader_pos.latitude = self._leader_home.latitude
        self._leader_pos.longitude = self._leader_home.longitude
        self._leader_pos.altitude = self._leader_home.altitude

        # set leader target
        self._leader_target.latitude = self._leader_home.latitude
        self._leader_target.longitude = self._leader_home.longitude
        self._leader_target.altitude = self._leader_home.altitude

    def _drone_diff(self):
        self._alt_diff = [0]

        for num in range(self._Drone_Num):
            if self._d_inDrCtrl[num].state.data in ["Armed", "Takeoff", "In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"]:
                self._drone_diff_pos[num].latitude = self._d_inDrCtrl[num].droneGPS.latitude - self._leader_pos.latitude
                self._drone_diff_pos[num].longitude = self._d_inDrCtrl[num].droneGPS.longitude - self._leader_pos.longitude
                if self._d_inDrCtrl[num].droneGPS.altitude < self._MinAlt[0]:
                    self._drone_diff_pos[num].altitude = self._takeoff_alt[num] - self._leader_pos.altitude
                else:
                    self._drone_diff_pos[num].altitude = self._d_inDrCtrl[num].droneGPS.altitude - self._leader_pos.altitude

                self._alt_diff.append(self._drone_diff_pos[num].altitude)

    def _diff_target_angle(self, target, pos):
        angle_vec_x = math.cos(math.radians(self._avg_mag))
        angle_vec_y = math.sin(math.radians(self._avg_mag))

        target_vec_x = target.latitude - pos.latitude
        target_vec_y = target.longitude - pos.longitude

        inner_product = angle_vec_x * target_vec_x + angle_vec_y * target_vec_y
        cross_product = angle_vec_x * target_vec_y - angle_vec_y * target_vec_x

        angle_rad = math.atan2(cross_product, inner_product)
        angle_deg = math.degrees(angle_rad)

        return angle_deg

    def _select_pivot(self):
        if self._Drone_Num > 1:
            dist_to_leader = [[0 for i in range(2)] for j in range(self._Drone_Num)]

            # 仮想リーダー機とドローン間の距離を求める
            for num in range(self._Drone_Num):
                dist_cal = drone_distance_cal.DroneMathmetics()
                dist_to_leader[num][0] = dist_cal.distance_math(self._leader_diff_pos.longitude,
                                              self._drone_diff_pos[num].longitude,
                                              self._leader_diff_pos.latitude,
                                              self._drone_diff_pos[num].latitude,
                                              0,
                                              0)
                dist_to_leader[num][1] = num

            # 全てのドローンと仮想リーダー機の距離が等しい場合は、リーダー機の機首方向に対して遠い機体を選ぶ処理
            # 以下コメントアウトした理由として、
            # 現実に全てのドローンと仮想リーダー機の距離が等しくなるケースは稀であり、また、
            # 以下の処理がなくともピボット機の選出は問題なくできるため。
            """
            # 仮想リーダー機と各ドローン間の距離がすべて等しいか調べる
            equal_dist = True

            for num in range(1, self._Drone_Num):
                if dist_to_leader[num-1][0] != dist_to_leader[num][0]:
                    equal_dist = False
            
            # 仮想リーダー機と各ドローン間の距離がすべて等しい場合は、仮想リーダー機のx軸に対して遠い2機をピボット機とする。
            # 上記以外の場合は、仮想リーダー機から距離の遠い2機をピボット機とする。
            if equal_dist:
                for num in range(self._Drone_Num):
                    deg = self._diff_target_angle(self._drone_diff_pos[num], self._leader_diff_pos)
                    dist_to_leader[num][0] = abs(math.sin(math.radians(deg)))
                    dist_to_leader[num][1] = num
            """
            # 求めた仮想リーダー機と各ドローンの距離から、遠い2機を選ぶ
            two_dist_drone = heapq.nlargest(2, dist_to_leader)
            self._pivot[0] = two_dist_drone[0][1]
            self._pivot[1] = two_dist_drone[1][1]

            logger.debug("pivot1:%d号機, pivot2:%d号機",(self._pivot[0]+1),(self._pivot[1]+1))

    def _decide_left_right_pivot(self):
        # 求めたピボット機が、左右どちらのピボット機であるか調べる
        angle_deg = self._diff_target_angle(self._drone_diff_pos[self._pivot[0]], self._leader_diff_pos)
        if angle_deg >= 0:
            self._right_pivot = self._pivot[0]
            self._left_pivot = self._pivot[1]
        else:
            self._right_pivot = self._pivot[1]
            self._left_pivot = self._pivot[0]

    def _set_target_gps(self):
        self._set_leader()
        self._config_target_gps()

        takeoff_landing = self._takeoff_landing()
        if takeoff_landing in [1, -1]:
            self._leader_pos = self._target_gps. \
                get_takeoff_landing_data(
                    self._leader_pos, takeoff_landing, self._leader_pos.altitude)
            for num in range(self._Drone_Num):
                if self._d_inDrCtrl[num].state.data == "Armed":
                    self._target[num] = self._target_gps. \
                        get_takeoff_landing_data(
                            self._d_inDrCtrl[num].homeGPS, takeoff_landing, self._takeoff_alt[num])

            if takeoff_landing == -1:
                self._rtl_instruction = True
                # set rtl alt
                self._set_rtl_alt()

        else:
            mv = self._moving()
            self._set_leader_alt(mv.up_down)

            turn_dir = self._turn_dir()

            if mv.front_back != 0 or mv.right_left != 0:
                self._mode = TRANS_MODE
            elif turn_dir in [1, -1]:
                if self._mode != TURN_MODE:
                    if self._avg_speed < DRONE_SPEED_RANGE_TO_START_TURN:
                        self._mode = TURN_MODE

                        #pivot機の左右計算
                        self._decide_left_right_pivot()

                        if turn_dir == 1:
                            self._turn_pivot = self._right_pivot
                            self._turn_end = self._left_pivot
                        else:
                            self._turn_pivot = self._left_pivot
                            self._turn_end = self._right_pivot

                        self._turn_angle_sum = 0
                        self._leader_pos_first = self._leader_pos
                        self._drone_diff_first = [0] * self._Drone_Num
                        for num in range(self._Drone_Num):
                            self._drone_diff_first[num] = self._drone_diff_pos[num]
                    else:
                        self._mode = NO_MOVE_MODE
                else:
                    pass
            else:
                self._mode = NO_MOVE_MODE

            # translational
            if self._mode == TRANS_MODE:
                # leader position
                dist_cal = drone_distance_cal.DroneMathmetics()
                dist_leader_target = dist_cal.distance_math(
                    self._leader_pos.longitude,
                    self._leader_target.longitude,
                    self._leader_pos.latitude,
                    self._leader_target.latitude,
                    0,
                    0)
                if dist_leader_target > 0:
                    pos_diff_lat = self._leader_target.latitude - self._leader_pos.latitude
                    pos_diff_lon = self._leader_target.longitude - self._leader_pos.longitude

                    self._leader_pos.latitude += (pos_diff_lat * ((self._avg_speed / self._execute_rate) / dist_leader_target))
                    self._leader_pos.longitude += (pos_diff_lon * ((self._avg_speed / self._execute_rate) / dist_leader_target))
                    self._leader_pos.altitude = self._leader_target.altitude

                target_update = self._update_target_gps(mv.front_back, mv.right_left, mv.up_down)
                if target_update:
                    # leader target position
                    self._leader_target = self._target_gps.get_move_target_data(
                        self._leader_pos, mv.front_back, mv.right_left, mv.up_down,
                        self._SpeedParam, self._move_mag)
                    # drone target position
                    for num in range(self._Drone_Num):
                        if self._d_inDrCtrl[num].state.data in ["In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"]:
                            self._target_gps.get_target_cal(self._target[num], self._leader_target, self._drone_diff_pos[num])

            # turn
            elif self._mode == TURN_MODE:
                # 旋回時にターゲット座標を更新するかを判定
                update = True
                for num in range(self._Drone_Num):
                    if self._d_inDrCtrl[num].state.data in ["In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"]:
                        dist_cal = drone_distance_cal.DroneMathmetics()
                        target_dist = dist_cal.distance_math(
                            self._d_inDrCtrl[num].droneGPS.longitude,
                            self._target[num].longitude,
                            self._d_inDrCtrl[num].droneGPS.latitude,
                            self._target[num].latitude,
                            0,
                            0)
                        if target_dist > TURN_UPDATE_DIST:
                            update = False

                # 次のターゲット座標を設定
                if update:
                    # leader position
                    self._leader_pos = self._leader_target

                    # target angular velocity
                    dist_cal = drone_distance_cal.DroneMathmetics()
                    dist = dist_cal.distance_math(
                        self._drone_diff_first[self._right_pivot].longitude,
                        self._drone_diff_first[self._left_pivot].longitude,
                        self._drone_diff_first[self._right_pivot].latitude,
                        self._drone_diff_first[self._left_pivot].latitude,
                        0,
                        0)

                    if dist == 0:
                        self._turn_angle = 0
                    else:
                        self._turn_angle = (180.0 * self._SpeedParam) / (math.pi * dist)

                    self._turn_angle_sum += self._turn_angle
                    if self._turn_angle_sum > 360:
                        self._turn_angle_sum = self._turn_angle_sum % 360
                    elif self._turn_angle_sum < 0:
                        self._turn_angle_sum = 360 - (abs(self._turn_angle_sum) % 360)

                    # distance of lat and lon 1 degree
                    self._target_gps._calc_lat_lon_degree_dist(self._leader_pos.latitude)
                    # leader target position
                    self._leader_diff_pos.altitude = self._leader_pos.altitude
                    self._leader_target = self._target_gps.get_turn_target_data(
                        self._drone_diff_first[self._turn_pivot], self._leader_diff_pos,
                        turn_dir, mv.up_down, self._turn_angle_sum)
                    self._leader_target.latitude += self._leader_pos_first.latitude
                    self._leader_target.longitude += self._leader_pos_first.longitude
                    # drone target position
                    for num in range(self._Drone_Num):
                        self._drone_diff_pos[num] = self._target_gps.get_turn_target_data(
                            self._leader_diff_pos, self._drone_diff_first[num],
                            turn_dir, 0, self._turn_angle_sum)

                        self._drone_diff_pos[num].altitude = self._drone_diff_first[num].altitude

                        self._target_gps.get_target_cal(
                            self._target[num], self._leader_target, self._drone_diff_pos[num])

            # hovering or cancel operation or home
            else:
                # leader position
                leader_lat_h = 0
                leader_lon_h = 0
                drone_num = 0
                for num in range(self._Drone_Num):
                    if self._d_inDrCtrl[num].state.data in ["In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"]:
                        leader_lat_h += (self._d_inDrCtrl[num].droneGPS.latitude - self._drone_diff_pos[num].latitude)
                        leader_lon_h += (self._d_inDrCtrl[num].droneGPS.longitude - self._drone_diff_pos[num].longitude)
                        drone_num += 1

                if drone_num > 0:
                    self._leader_pos.latitude = (leader_lat_h/drone_num)
                    self._leader_pos.longitude = (leader_lon_h/drone_num)
                    self._leader_pos.altitude = self._leader_target.altitude

                # leader target position
                self._leader_target = self._target_gps.get_move_target_data(
                    self._leader_pos, mv.front_back, mv.right_left, mv.up_down,
                    self._SpeedParam, self._avg_mag)
                # drone target position
                for num in range(self._Drone_Num):
                    if self._d_inDrCtrl[num].state.data in ["Armed", "Takeoff", "In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"]:
                        self._target_gps.get_target_cal(
                            self._target[num], self._leader_target, self._drone_diff_pos[num])

            self._last_front_back = mv.front_back
            self._last_right_left = mv.right_left
            self._last_up_down = mv.up_down

        # distance between drone
        for pos1 in range(self._Drone_Num):
            for pos2 in range((pos1 + 1), self._Drone_Num):
                if (self._d_inDrCtrl[pos1].state.data in ["In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"] and
                    self._d_inDrCtrl[pos2].state.data in ["In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"]):

                    dist_cal = drone_distance_cal.DroneMathmetics()
                    dist = dist_cal.distance_math(
                        self._d_outTarget[pos1].pos.longitude,
                        self._d_outTarget[pos2].pos.longitude,
                        self._d_outTarget[pos1].pos.latitude,
                        self._d_outTarget[pos2].pos.latitude,
                        0,
                        0)
                    logger.debug("drone%d-drone%d_dist: % f",pos1+1, pos2+1, dist)

        # set outTarget
        for num in range(self._Drone_Num):
            if self._rtl_instruction:
                self._d_outTarget[num].pos.latitude = self._d_inDrCtrl[num].homeGPS.latitude
                self._d_outTarget[num].pos.longitude = self._d_inDrCtrl[num].homeGPS.longitude
                self._d_outTarget[num].pos.altitude = self._d_inDrCtrl[num].homeGPS.altitude
            elif self._d_inDrCtrl[num].state.data in ["RTL", "Disarm"]:
                self._d_outTarget[num].pos.latitude = self._d_inDrCtrl[num].homeGPS.latitude
                self._d_outTarget[num].pos.longitude = self._d_inDrCtrl[num].homeGPS.longitude
                self._d_outTarget[num].pos.altitude = self._d_inDrCtrl[num].homeGPS.altitude
            else:
                self._d_outTarget[num].pos.latitude = self._target[num].latitude
                self._d_outTarget[num].pos.longitude = self._target[num].longitude
                self._d_outTarget[num].pos.altitude = self._target[num].altitude

    def _set_leader_alt(self, up_down):
        if up_down:
            #仮想リーダー機は高度変更指示に対してタイムリーに高度を変更することができるが、
            #実際のドローンはターゲットの高度に到達するまで時間がかかる
            #実際のドローンが仮想リーダー機の±20cmに接近した時に仮想リーダー機の高度を変更するための処理
            #以下の判断がないと仮想リーダー機の高度が上がり続けてしまうので注意
            diff_avg = 0
            for num in range(self._Drone_Num):
                if self._d_inDrCtrl[num].state.data in ["Armed", "Takeoff", "In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"]:
                    diff_avg += self._drone_diff_pos[num].altitude

            diff_avg /= (self._Drone_Num - self._rtl_num)

            if abs(self._leader_target.altitude - (self._avg_drone_pos.altitude - diff_avg)) > LEADER_ALTITUDE_CHANGE:
                self._leader_target.altitude = self._leader_pos.altitude

        else:
            # after alt change set avg_drone_altitude to leader_altitude for 3 seconds
            if self._last_up_down in [1, -1]:
                self._alt_adjust_cnt = 0
            if 0 <= self._alt_adjust_cnt < (self._execute_rate * ALT_ADJUST_TIME):
                self._alt_adjust_cnt += 1
                self._leader_target.altitude = self._avg_drone_pos.altitude

    def _update_target_gps(self, front_back, right_left, up_down):
        target_update = True

        move_continue = self._move_continue(front_back, right_left, up_down)
        if move_continue:
            dist_cal = drone_distance_cal.DroneMathmetics()
            dist_leader_target = dist_cal.distance_math(
                self._leader_pos.longitude,
                self._leader_target.longitude,
                self._leader_pos.latitude,
                self._leader_target.latitude,
                0,
                0)

            if self._d_outMoveMode.data == "Absolute":
                if front_back:
                    update_coefficient = MOVE_UPDATE_COEF_ABSOLUTE_NS
                else:
                    update_coefficient = MOVE_UPDATE_COEF_ABSOLUTE_EW
            else:
                update_coefficient = MOVE_UPDATE_COEF_RELATIVE

            if self._last_speedparam == self._SpeedParam:
                if dist_leader_target >= (self._SpeedParam * self._LatLonCoefficient[0] * update_coefficient):
                    target_update = False
        else:
            self._move_mag = self._avg_mag

        return target_update

    def _set_target_speed(self):
        self._d_outSettingSpeed.data = self._SpeedParam
        if self._mode == TRANS_MODE:
            self._trans_target_speed()
        elif self._mode == TURN_MODE:
            self._turn_target_speed()

    def _trans_target_speed(self):
        # leader next target speed
        leader_next_speed = self._SpeedParam

        # drone target speed
        dist_cal = drone_distance_cal.DroneMathmetics()
        dist_leader_target = dist_cal.distance_math(self._leader_pos.longitude,
                                                    self._leader_target.longitude,
                                                    self._leader_pos.latitude,
                                                    self._leader_target.latitude,
                                                    0,
                                                    0)

        for num in range(self._Drone_Num):
            if self._d_inDrCtrl[num].state.data in ["In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"]:
                dist_drone_target = dist_cal.distance_math(
                    self._d_inDrCtrl[num].droneGPS.longitude,
                    self._d_outTarget[num].pos.longitude,
                    self._d_inDrCtrl[num].droneGPS.latitude,
                    self._d_outTarget[num].pos.latitude,
                    0,
                    0)
                if dist_leader_target == 0:
                    drone_target_speed = leader_next_speed
                else:
                    drone_target_speed = leader_next_speed * (dist_drone_target / dist_leader_target)

                self._d_outTarget[num].speed.data = drone_target_speed
 
    def _turn_target_speed(self):
        dist_cal = drone_distance_cal.DroneMathmetics()
        dist = dist_cal.distance_math(
            self._drone_diff_pos[self._right_pivot].longitude,
            self._drone_diff_pos[self._left_pivot].longitude,
            self._drone_diff_pos[self._right_pivot].latitude,
            self._drone_diff_pos[self._left_pivot].latitude,
            0,
            0)

        # drone target speed
        leader_real_speed = self._SpeedParam
        dist_cal = drone_distance_cal.DroneMathmetics()
        for num in range(self._Drone_Num):
            if self._d_inDrCtrl[num].state.data in ["In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"]:
                rotation_axis_dist = dist_cal.distance_math(
                    self._drone_diff_pos[self._turn_pivot].longitude,
                    self._drone_diff_pos[num].longitude,
                    self._drone_diff_pos[self._turn_pivot].latitude,
                    self._drone_diff_pos[num].latitude,
                    0,
                    0)

                if dist == 0:
                    drone_target_speed = 0
                else:
                    drone_target_speed = leader_real_speed * (rotation_axis_dist / dist)

                self._d_outTarget[num].speed.data = drone_target_speed

    def _set_target_yaw(self):
        angle_deg = 0
        target_deg = 0
        target_dir = 0

        # drone target angle
        if self._mode == TRANS_MODE:
            if self._d_outMoveMode.data == "Absolute":
                angle_deg = self._diff_target_angle(self._leader_target, self._leader_pos)
                if angle_deg > 0:
                    target_dir = 1
                else:
                    target_dir = -1

                target_deg = int(self._avg_mag + angle_deg)

            elif self._d_outMoveMode.data == "Relative":
                    target_dir = 1
                    target_deg = int(self._move_mag)

        elif self._mode == TURN_MODE:
            target_dir = int(self._turn_dir())

            if self._right_pivot == self._left_pivot:
                angle_deg = ONLY_ONE_TURN_SPEED * target_dir
            else:
                angle_deg = self._diff_target_angle(self._leader_target, self._leader_pos)

            target_deg = int(self._avg_mag + angle_deg)

        if target_deg > 360:
            target_deg = target_deg % 360
        elif target_deg < 0:
            target_deg = 360 - (abs(target_deg) % 360)

        for num in range(self._Drone_Num):
            if self._d_inDrCtrl[num].state.data in ["In_Move", "Alt_CHG", "Hovering", "Yaw_CHG"]:
                self._d_outTarget[num].yaw.data = [target_dir, target_deg]

    def _set_move_mode(self):
        mode = self._move_mode()

        if mode == 1:
            self._d_outMoveMode.data = "Absolute"
        else:
            self._d_outMoveMode.data = "Relative"

        self._target_gps.set_controller_mode(mode)

    def _set_leader_target_gps(self):

        if self._rtl_instruction:
            # set home position when after RTL instruction 
            self._d_outVirtualLeader.droneGPS.latitude = self._leader_home.latitude
            self._d_outVirtualLeader.droneGPS.longitude = self._leader_home.longitude
            self._d_outVirtualLeader.droneGPS.altitude = self._leader_home.altitude
            self._d_outVirtualLeader.targetGPS.latitude = self._leader_home.latitude
            self._d_outVirtualLeader.targetGPS.longitude = self._leader_home.longitude
            self._d_outVirtualLeader.targetGPS.altitude = self._leader_home.altitude
            self._d_outVirtualLeader.mag.data = self._avg_mag
        else:
            for num in range(self._Drone_Num):
                if self._d_inDrCtrl[num].state.data in ["RTL", "Disarm"]:
                    all_rtl = True
                else:
                    all_rtl = False
                    break
            if all_rtl:
                # set home position when all drone is RTL state
                self._d_outVirtualLeader.droneGPS.latitude = self._leader_home.latitude
                self._d_outVirtualLeader.droneGPS.longitude = self._leader_home.longitude
                self._d_outVirtualLeader.droneGPS.altitude = self._leader_home.altitude
                self._d_outVirtualLeader.targetGPS.latitude = self._leader_home.latitude
                self._d_outVirtualLeader.targetGPS.longitude = self._leader_home.longitude
                self._d_outVirtualLeader.targetGPS.altitude = self._leader_home.altitude
                self._d_outVirtualLeader.mag.data = self._avg_mag
            else:
                # set leader gps until the all drone is RTL state
                self._d_outVirtualLeader.droneGPS.latitude = self._leader_pos.latitude
                self._d_outVirtualLeader.droneGPS.longitude = self._leader_pos.longitude
                self._d_outVirtualLeader.droneGPS.altitude = self._leader_pos.altitude
                self._d_outVirtualLeader.targetGPS.latitude = self._leader_target.latitude
                self._d_outVirtualLeader.targetGPS.longitude = self._leader_target.longitude
                self._d_outVirtualLeader.targetGPS.altitude = self._leader_target.altitude
                self._d_outVirtualLeader.mag.data = self._avg_mag

    def _is_drone_on_home(self):
        home = False

        dist = [None] * self._Drone_Num
        dist_cal = drone_distance_cal.DroneMathmetics()
        for num in range(self._Drone_Num):
            dist[num] = dist_cal.distance_math(
                self._d_inDrCtrl[num].homeGPS.longitude,
                self._d_inDrCtrl[num].droneGPS.longitude,
                self._d_inDrCtrl[num].homeGPS.latitude,
                self._d_inDrCtrl[num].droneGPS.latitude,
                0,
                0)

            if self._rtl_instruction:
                # 全機体がhome上空に到達したか
                if dist[num] <= IN_HOME_AREA:
                    home = True
                else:
                    home = False
                    break
            else:
                # 全機体がhome領域内にいるかどうか
                if (dist[num] <= ALLOWABLE_ERROR_POS_CMP and
                    self._d_inDrCtrl[num].droneGPS.altitude < (self._MinAlt[0] - ALLOWABLE_ERROR_HEIGHT)):
                    home = True
                else:
                    home = False
                    break

        self._home = home

    def _takeoff_landing(self):
        operation = 0
        if self._home:
            # check takeoff request
            if self._d_ButtonIn.data & BUTTON_06_MASK:
                operation = 1
            elif self._d_ButtonIn.data & BUTTON_08_MASK:
                logger.warning("WARNING: Landing operation is "
                               "NOT available "
                               "in the current drone position")
        else:
            # check landing request
            if self._d_ButtonIn.data & BUTTON_08_MASK:
                operation = -1
            elif self._d_ButtonIn.data & BUTTON_06_MASK:
                logger.warning("WARNING: Take-off operation is "
                               "NOT available in the current drone position")

        return operation

    def _move_mode(self):
        mode = self._target_gps._mode

        if self._d_ButtonIn.data & BUTTON_10_MASK:
            if mode == 1:
                mode = 2
            else:
                mode = 1

        return mode

    def _moving(self):
        mv1 = self._front_back()
        mv2 = self._right_left()
        mv3 = self._up_down()

        if self._home:
            if mv1 or mv2 or mv3:
                mv1 = mv2 = mv3 = 0
                logger.warning("WARNING: Move operations are "
                               "NOT available in the current drone position")

        return self.MovingOperation(mv1, mv2, mv3)

    def _front_back(self):
        key = 0
        try:
            key = self._d_AnalogIn.data[0]
        except IndexError:
            pass

        operation = self._front_back_check(key)
        operation *= -1

        return operation

    def _right_left(self):
        operation = 0
        both_push = BUTTON_03_MASK | BUTTON_01_MASK
        if (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [right/left]")
        elif self._d_ButtonIn.data & BUTTON_03_MASK:
            operation = 1
        elif self._d_ButtonIn.data & BUTTON_01_MASK:
            operation = -1

        return operation

    def _up_down(self):
        operation = 0
        both_push = BUTTON_04_MASK | BUTTON_02_MASK
        if (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [up/down]")
        elif self._d_ButtonIn.data & BUTTON_04_MASK:
            operation = 1
        elif self._d_ButtonIn.data & BUTTON_02_MASK:
            operation = -1

        return operation

    def _move_continue(self, front_back, right_left, up_down):
        if self._last_move_mode == self._d_outMoveMode.data:
            if (front_back != self._last_front_back) or (right_left != self._last_right_left) or (up_down != self._last_up_down):
                move_continue = False
            elif (front_back == 0) and (right_left == 0) and (up_down == 0):
                move_continue = False
            else:
                move_continue = True
        else:
            move_continue = False

        return move_continue

    def _turn_dir(self):
        key = 0
        home = False
        for num in range(self._Drone_Num):
            if self._d_inDrCtrl[num].state.data in ["Armed", "Takeoff"]:
                home = True

        try:
            key = self._d_AnalogIn.data[1]
        except IndexError:
            pass

        if home:
            if key:
                key = 0
                logger.warning("WARNING: Turning operation is "
                               "NOT available in the current drone position")

        else:
            if self._d_ButtonIn.data & BUTTON_08_MASK:
                key = 0
            else:
                if self._d_outMoveMode.data == "Absolute":
                    mv1 = self._front_back()
                    mv2 = self._right_left()
                    if mv1 or mv2:
                        key = 0

        operation = self._turn_dir_check(key)

        return operation

    def _move_speed(self):
        both_push = BUTTON_05_MASK | BUTTON_07_MASK

        if self._home:
            if (self._d_ButtonIn.data & BUTTON_05_MASK) or (self._d_ButtonIn.data & BUTTON_07_MASK) or (self._d_ButtonIn.data & both_push):
                logger.warning("WARNING: Speed operation is "
                               "NOT available in the current drone position")
        elif (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [MoveSpeed up/down]")
        elif self._d_ButtonIn.data & BUTTON_05_MASK:
            self._SpeedParam += 1.0

        elif self._d_ButtonIn.data & BUTTON_07_MASK:
            self._SpeedParam -= 1.0

        self._SpeedParam = self._move_speed_check(self._SpeedParam)

    def _front_back_check(self, status):
        if status > 1:
            status = 1
        elif status < -1:
            status = -1

        return status

    def _turn_dir_check(self, dir):
        if dir > 1:
            dir = 1
        elif dir < -1:
            dir = -1

        return dir

    def _move_speed_check(self, speed):
        if speed > MOVE_SPEED_MAX:
            speed = MOVE_SPEED_MAX
        elif speed < MOVE_SPEED_MIN:
            speed = MOVE_SPEED_MIN

        return speed

    def _is_landing(self):
        landing = False
        for num in range(self._Drone_Num):
            if self._d_inDrCtrl[num].state.data == "RTL":
                landing = True

        return landing

    def _takeoff_range_check(self):
        ret = True
        self._takeoff_alt = [None] * INPORT_NUM
        self._takeoff_alt[0] = self._Drone1TakeoffAlt[0]
        self._takeoff_alt[1] = self._Drone2TakeoffAlt[0]
        self._takeoff_alt[2] = self._Drone3TakeoffAlt[0]

        for num in range(self._Drone_Num):
            if (self._takeoff_alt[num] < self._MinAlt[0]) or (self._takeoff_alt[num] > self._MaxAlt[0]):
                logger.error("ERROR: Takeoff altitude is out of range")
                ret = False
                break

        return ret

    def _set_rtl_alt(self):
        self._getRTLAlt = [None] * SV_RTLALT_NUM
        self._getRTLAlt[0] = self._getRTLAlt1
        self._getRTLAlt[1] = self._getRTLAlt2
        self._getRTLAlt[2] = self._getRTLAlt3

        self._rtl_alt = [None] * self._Drone_Num
        for num in range(self._Drone_Num):
            self._rtl_alt[num] = self._getRTLAlt[num]._ptr().get_rtl_alt()

    # 入力されたGameControllerのボタン情報が有効な値かどうかを調べる
    def _is_valid_button(self, button):
        ret = False
        if 0 <= button < pow(2, BUTTON_NUM):
            ret = True
        else:
            logger.debug("invalid Button = 0b" + format(button, '016b'))

        return ret

    # 入力されたGameControllerのアナログ情報が有効な値かどうかを調べる
    def _is_valid_analog(self, analog, key_num):
        ret = False
        if key_num in [ANALOG_DIR_UP_DOWN, ANALOG_DIR_LIGHT_LEFT]:
            if analog in [-1, 0, 1]:
                ret = True
        elif key_num in [ANALOG_LSTICK_UP_DOWN, ANALOG_LSTICK_LIGHT_LEFT,
                        ANALOG_RSTICK_UP_DOWN, ANALOG_RSTICK_LIGHT_LEFT,
                        ANALOG_L2, ANALOG_R2]:
            if -1 <= analog <= 1:
                ret = True
        else:
            ret = True

        if ret == False:
            logger.debug("invalid Analog[%d]:  % f", key_num, analog)

        return ret

    # 入力されたDroneIOからの情報が有効な値かどうかを調べる
    def _is_valid_drctrl(self, drctrl, num):
        ret = False

        # drctrl check
        Id = self._is_valid_id(drctrl.id, num, "Rcv ID")
        home_lat = self._is_valid_gps_lat(drctrl.homeGPS.latitude, num, "HomeGPS")
        home_lon = self._is_valid_gps_lon(drctrl.homeGPS.longitude, num, "HomeGPS")
        drone_lat = self._is_valid_gps_lat(drctrl.droneGPS.latitude, num, "DroneGPS")
        drone_lon = self._is_valid_gps_lon(drctrl.droneGPS.longitude, num, "DroneGPS")
        drone_alt = self._is_valid_gps_alt(drctrl.droneGPS.altitude, drctrl.state.data, num, "DroneGPS")
        mag = self._is_valid_angle(drctrl.mag.data, num, "Mag")
        state = self._is_valid_state(drctrl.state.data, num)

        if Id and home_lat and home_lon and drone_lat and drone_lon and drone_alt and mag and state:
            ret = True

        return ret

    # 出力するターゲット情報が有効な値かどうかを調べる
    def _is_valid_target(self, target, num):
        ret = False

        # target check
        Id = self._is_valid_id(target.id, num, "Send ID")
        target_lat = self._is_valid_gps_lat(target.pos.latitude, num, "TargetGPS")
        target_lon = self._is_valid_gps_lon(target.pos.longitude, num, "TargetGPS")
        target_alt = self._is_valid_gps_alt(target.pos.altitude, self._d_inDrCtrl[num].state.data, num, "TargetGPS")
        target_speed = self._is_valid_speed(target.speed.data, num)
        target_yaw_dir = self._is_valid_angle_dir(target.yaw.data[0], num, "TargetYaw_dir")
        target_yaw_angle = self._is_valid_angle(target.yaw.data[1], num, "TargetYaw_angle")

        if Id and target_lat and target_lon and target_alt and target_speed and target_yaw_dir and target_yaw_angle:
            ret = True

        return ret

    # 出力する仮想リーダー機の位置情報が有効な値かどうかを調べる
    def _is_valid_virtual_leader(self, virtual_leader):
        ret = False

        # target check
        drone_lat = self._is_valid_gps_lat(virtual_leader.droneGPS.latitude, -1, "Virtual_Leader_DroneGPS")
        drone_lon = self._is_valid_gps_lon(virtual_leader.droneGPS.longitude, -1, "Virtual_Leader_DroneGPS")
        target_lat = self._is_valid_gps_lat(virtual_leader.targetGPS.latitude, -1, "Virtual_Leader_TargetGPS")
        target_lon = self._is_valid_gps_lon(virtual_leader.targetGPS.longitude, -1, "Virtual_Leader_TargetGPS")
        mag = self._is_valid_angle(virtual_leader.mag.data, -1, "Virtual_Leader_Mag")

        if drone_lat and drone_lon and target_lat and target_lon and mag:
            ret = True

        return ret

    # idが有効値かチェックする
    def _is_valid_id(self, Id, num, name):
        ret = False
        if Id >= 0:
            ret = True
        else:
            logger.debug("invalid %s[%d]:  % d", name, num, Id)

        return ret

    # 緯度値が有効値かチェックする
    def _is_valid_gps_lat(self, lat, num, name):
        ret = False
        if -90.0 <= lat <= 90.0:
            ret = True
        else:
            logger.debug("invalid %s[%d] lat:  % f", name, num, lat)

        return ret

    # 経度値が有効値かチェックする
    def _is_valid_gps_lon(self, lon, num, name):
        ret = False
        if -180.0 <= lon <= 180.0:
            ret = True
        else:
            logger.debug("invalid %s[%d] lon:  % f", name, num, lon)

        return ret

    # 高度値が有効値かチェックする
    def _is_valid_gps_alt(self, alt, state, num, name):
        ret = False
        if state in ["Hovering","In_Move","Alt_CHG","Yaw_CHG"]:
            if self._MinAlt[0] <= alt <= self._MaxAlt[0]:
                ret =  True
            else:
                logger.debug("invalid %s[%d] alt:  % f", name, num, alt)
        else:
            ret = True

        return ret

    # 旋回角度が有効値かチェックする
    def _is_valid_angle(self, angle, num, name):
        ret = False
        if 0 <= angle <= 360:
            ret = True
        else:
            logger.debug("invalid %s[%d]:  % d", name, num, angle)

        return ret

    # 旋回方向の値が有効値かチェックする
    def _is_valid_angle_dir(self, angle_dir, num, name):
        ret = False
        if angle_dir in [-1, 0, 1]:
            ret = True
        else:
            logger.debug("invalid %s[%d]:  % d", name, num, angle_dir)

        return ret

    # DroneIO内部ステータスが有効値かチェックする
    def _is_valid_state(self, state, num):
        ret = False
        if state in ["Armed","Takeoff","Hovering","In_Move","Alt_CHG","Yaw_CHG","RTL","Disarm"]:
            ret = True
        else:
            logger.debug("invalid State[%d]:  % s", num, state)

        return ret

    # 速度が有効値かチェックする
    def _is_valid_speed(self, speed, num):
        ret = False
        if speed >= 0:
            ret = True
        else:
            logger.debug("invalid TargetSpeed[%d]:  % f", num, speed)

        return ret

    def _cb_connect_status_change(self, port, connect):
        if port == "ButtonIn":
            self._button = connect
        elif port == "AnalogIn":
            self._analog = connect
        else:
            logger.warning("Invalid port name: %s", port)

        if not connect:
            logger.warning("ERROR: Data port was disconnected: %s", port)

    def _ready_to_log_write(self, file_mode):
        
        if file_mode == FILE_MODE_WRITE:
            date, time = self._get_datetime_to_str()
            time = time.replace(":", "").replace(".", "")  #  HH:mm:ss.sss --> HHmmssSSS
            self._file_name = date + "_" + time[0:6] + "_ctrl_log.csv"

        if self._is_open_rtc_viewer_log(file_mode):
            if self._is_csv_writer():
                if file_mode == FILE_MODE_WRITE:
                    self._writer.writerow(self._header)
        
    def _get_datetime_to_str(self):
    
        curr_time = datetime.now()
        date = "%s%s%s" % ("{0:04d}".format(curr_time.year), "{0:02d}".format(curr_time.month), "{0:02d}".format(curr_time.day))
        time = "%s:%s:%s.%s" % ("{0:02d}".format(curr_time.hour), "{0:02d}".format(curr_time.minute),
                           "{0:02d}".format(curr_time.second), "{0:03d}".format(curr_time.microsecond / 1000))
        return date, time

    def _is_open_rtc_viewer_log(self, file_mode):

        rtn = False
        
        try:
            if self._file_name != "":
                self._file = open(self._file_name, file_mode)
                rtn = True
        except IOError as e:
            logger.error("I/O Error {0} {1}".format(e.errno, e.strerror))

        return rtn
                
    def _is_csv_writer(self):

        rtn = False

        try:
            if self._file is not None:
                self._writer = csv.writer(self._file)
                if self._writer is not None:
                    rtn = True
        except csv.Error as e:
            logger.error("CSV writer error {0} {1}".format(e.errno, e.strerror))

        return rtn

    def _write_rtc_viewer_log(self, rtc_state):

        if self._writer is not None:
            self._set_log(rtc_state)
            self._write_log(self._log_row, rtc_state)
    
    def _close_rtc_viewer_log(self):

        if self._file is not None:
            self._file.close()
            self._file = None
            self._writer = None

    def _conv_rtctime_to_datetime(self, tm):

        curr_time = datetime.fromtimestamp(tm.sec)

        date = "%s%s%s" % ("{0:04d}".format(curr_time.year), "{0:02d}".format(curr_time.month),
                           "{0:02d}".format(curr_time.day))
        time = "%s:%s:%s.%s" % ("{0:02d}".format(curr_time.hour), "{0:02d}".format(curr_time.minute),
                                "{0:02d}".format(curr_time.second), "{0:03d}".format(tm.nsec / 1000000))

        return date, time

    def _set_log(self, rtc_state):
    
        if rtc_state == "Active":
            ele_num = self._Drone_Num
        else:
            ele_num = 1

        for num in range(ele_num):
            self._set_rcv_log(num, rtc_state)
            self._set_snd_log(num, rtc_state)
            self._set_other_log(num, rtc_state)

    def _set_rcv_log(self, num, rtc_state):

        if rtc_state == "Active":
            if self._dr_ctrl_rcv[num]:
                rcv_date, rcv_time = self._conv_rtctime_to_datetime(self._d_inDrCtrl[num].tm)

                self._log_row[num][self._header.index("RcvDate")] = rcv_date
                self._log_row[num][self._header.index("RcvTime")] = rcv_time
                self._log_row[num][self._header.index("RcvRobotID")] = self._d_inDrCtrl[num].robotID.data
                self._log_row[num][self._header.index("RcvID")] = self._d_inDrCtrl[num].id
                self._log_row[num][self._header.index("HomeLat")] = "{0:.10f}".format(self._d_inDrCtrl[num].homeGPS.latitude)
                self._log_row[num][self._header.index("HomeLon")] = "{0:.10f}".format(self._d_inDrCtrl[num].homeGPS.longitude)
                self._log_row[num][self._header.index("HomeAlt")] = "{0:.2f}".format(self._d_inDrCtrl[num].homeGPS.altitude)
                self._log_row[num][self._header.index("RobotLat")] = "{0:.10f}".format(self._d_inDrCtrl[num].droneGPS.latitude)
                self._log_row[num][self._header.index("RobotLon")] = "{0:.10f}".format(self._d_inDrCtrl[num].droneGPS.longitude)
                self._log_row[num][self._header.index("RobotAlt")] = "{0:.2f}".format(self._d_inDrCtrl[num].droneGPS.altitude)
                self._log_row[num][self._header.index("AccX")] = "{0:.2f}".format(self._d_inDrCtrl[num].accel.data[0])
                self._log_row[num][self._header.index("AccY")] = "{0:.2f}".format(self._d_inDrCtrl[num].accel.data[1])
                self._log_row[num][self._header.index("AccZ")] = "{0:.2f}".format(self._d_inDrCtrl[num].accel.data[2])
                self._log_row[num][self._header.index("Mag")] = int(self._d_inDrCtrl[num].mag.data)
                self._log_row[num][self._header.index("State")] = self._d_inDrCtrl[num].state.data

    def _set_snd_log(self, num, rtc_state):

        if rtc_state == "Active":
            snd_date, snd_time = self._conv_rtctime_to_datetime(self._d_outMoveMode.tm)
            self._log_row[num][self._header.index("SndDate")] = snd_date
            self._log_row[num][self._header.index("SndTime")] = snd_time
            self._log_row[num][self._header.index("MoveMode")] = self._d_outMoveMode.data

            if self._out_target_wrote[num]:
                snd_date, snd_time = self._conv_rtctime_to_datetime(self._d_outTarget[num].tm)
                self._log_row[num][self._header.index("SndDate")] = snd_date
                self._log_row[num][self._header.index("SndTime")] = snd_time
                self._log_row[num][self._header.index("SndRobotID")] = self._d_inDrCtrl[num].robotID.data
                self._log_row[num][self._header.index("SndID")] = self._d_outTarget[num].id
                self._log_row[num][self._header.index("TargetLat")] = "{0:.10f}".format(self._d_outTarget[num].pos.latitude)
                self._log_row[num][self._header.index("TargetLon")] = "{0:.10f}".format(self._d_outTarget[num].pos.longitude)
                self._log_row[num][self._header.index("TargetAlt")] = "{0:.2f}".format(self._d_outTarget[num].pos.altitude)
                self._log_row[num][self._header.index("TargetSpd")] = "{0:.2f}".format(self._d_outTarget[num].speed.data)
                self._log_row[num][self._header.index("TargetYawDir")] = self._d_outTarget[num].yaw.data[0]
                self._log_row[num][self._header.index("TargetYawDeg")] = self._d_outTarget[num].yaw.data[1]

    def _set_other_log(self, num, rtc_state):

        if rtc_state == "Active":
            self._log_row[num][self._header.index("Category")] = "GCS"
            self._log_row[num][self._header.index("ActStartDate")] = self._act_start_date
            self._log_row[num][self._header.index("ActStartTime")] = self._act_start_time
            self._log_row[num][self._header.index("RTCState")] = rtc_state
        else:
            self._log_row[self._header.index("Category")] = "GCS"
            self._log_row[self._header.index("ActStartDate")] = self._act_start_date
            self._log_row[self._header.index("ActStartTime")] = self._act_start_time
            self._log_row[self._header.index("RTCState")] = rtc_state

    def _write_log(self, log, rtc_state):

        if rtc_state == "Active":
            for num in range(len(log)):
                if all(item is "NULL" for item in log[num]) is False:
                    self._writer.writerow(log[num])
        else:
            self._writer.writerow(log)


    class MovingOperation(object):

        def __init__(self, front_back, right_left, up_down):
            self.front_back = front_back
            self.right_left = right_left
            self.up_down = up_down


class ConnectionCallback(OpenRTM_aist.PortCallBack.ConnectionCallback):

    def __init__(self, func):
        self._plugin_func = func

    def __call__(self, profile):
        self._plugin_func()


def DroneFlightDirectorInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=droneflightdirector_spec)
    manager.registerFactory(profile,
                            DroneFlightDirector,
                            OpenRTM_aist.Delete)


def MyModuleInit(manager):
    DroneFlightDirectorInit(manager)

    # Create a component
    comp = manager.createComponent("DroneFlightDirector")


def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager()

if __name__ == "__main__":
    main()
