#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file target_gps.py
 @brief This module makes drone target GPS data.
 @date $Date$

 @author East Japan Accounting Center Co., Ltd.

 (c)2017 The University of Aizu
 This software is released under the New BSD License.

"""
import sys
import RTC
from logging import getLogger, StreamHandler
from logging import DEBUG as LOG_LV
import math

sys.path.append(".")

# constants
CONTROL_MODE_1 = 1     #Absolute
CONTROL_MODE_2 = 2     #Relative
RIGHT_TURN_MODE = 1
LEFT_TURN_MODE = -1

EARTH_EQUATOR_DIST = 40075.0 #赤道の距離
EARTH_MERIDIAN_DIST = 40008.0 #子午線の距離

# Setup logger
logger = getLogger(__name__)
handler = StreamHandler()
handler.setLevel(LOG_LV)
logger.setLevel(LOG_LV)
logger.addHandler(handler)


class TargetGPS(object):
    def __init__(self):
        self._init_variable()

    def _init_variable(self):
        self._mode = CONTROL_MODE_1
        self._AltCoefficient = 1
        self._LatLonCoefficient = 10
        self._LatDiff10m = 0.000137
        self._LonDiff10m = 0.000108
        self._NoseLatDiff10m = 9.01323
        self._NoseLonDiff10m = 8.9831566
        self._MoveSpeed = 3.0
        self._MinAlt = 2.0
        self._MaxAlt = 15.0
        self._DroneNum = 3

    def config_param(self, kwargs):
        key1 = key2 = key3 = key4 = key5 = key6 = key7 = key8 = key9 = key10 = None
        try:
            key1 = kwargs["AltCoefficient"]
            key2 = kwargs["LatLonCoefficient"]
            key3 = kwargs["LatDiff10m"]
            key4 = kwargs["LonDiff10m"]
            key5 = kwargs["NoseLatDiff10m"]
            key6 = kwargs["NoseLonDiff10m"]
            key7 = kwargs["MoveSpeed"]
            key8 = kwargs["MinAlt"]
            key9 = kwargs["MaxAlt"]
            key10 = kwargs["DroneNum"]

        except KeyError:
            logger.warning("ERROR: KeyError")

        if key1 is not None:
            self._AltCoefficient = key1

        if key2 is not None:
            self._LatLonCoefficient = key2

        if key3 is not None:
            self._LatDiff10m = key3

        if key4 is not None:
            self._LonDiff10m = key4

        if key5 is not None:
            self._NoseLatDiff10m = key5

        if key6 is not None:
            self._NoseLonDiff10m = key6

        if key7 is not None:
            self._MoveSpeed = key7

        if key8 is not None:
            self._MinAlt = key8

        if key9 is not None:
            self._MaxAlt = key9

        if key10 is not None:
           self._DroneNum = key10

    def set_controller_mode(self, mode):
        if mode == 1:
            self._mode = CONTROL_MODE_1
        elif mode == 2:
            self._mode = CONTROL_MODE_2


    def get_takeoff_landing_data(self, home, takeoff_landing, alt):
        takeoff_landing = self._button_status_check(takeoff_landing)

        if takeoff_landing:
            target = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                 home.latitude, home.longitude, home.altitude,
                                 0.0, 0.0, 0.0, 0.0,
                                 0.0, 0, RTC.GPS_FIX_NORMAL)

            if takeoff_landing == 1:
                target.altitude = alt

        return target

    def get_move_target_data(self, drone, front_back, right_left, up_down, speedparam, mag):
        if self._mode == CONTROL_MODE_1:
            target = self._move_mode1(drone, front_back, right_left, up_down, speedparam)
        elif self._mode == CONTROL_MODE_2:
            target = self._move_mode2(drone, front_back, right_left, up_down, speedparam, mag)


        return target

    def get_turn_target_data(self, pivot, drone, turn_dir, up_down, mag):
        if turn_dir == RIGHT_TURN_MODE:
            target = self._right_turn_mode(pivot, drone, up_down, mag)
        elif turn_dir == LEFT_TURN_MODE:
            target = self._left_turn_mode(pivot, drone, up_down, mag)

        return target


    def _move_mode1(self, drone, south_north, east_west, up_down, speedparam):
        target = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0, RTC.GPS_FIX_NORMAL)

        target.latitude = self._move_south_north(drone.latitude, south_north, speedparam)
        target.longitude = self._move_east_west(drone.longitude, east_west, speedparam)
        target.altitude = self._move_up_down(drone.altitude, up_down)

        return target

    def _move_mode2(self, drone, front_back, right_left, up_down, speedparam, mag):
        target = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0, RTC.GPS_FIX_NORMAL)

        target.latitude,target.longitude = self._move_nose(drone, front_back, right_left, speedparam, mag)
        target.altitude = self._move_up_down(drone.altitude, up_down)

        return target

    def _right_turn_mode(self, pivot, drone, up_down, mag):
        target = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0, RTC.GPS_FIX_NORMAL)

        target.latitude = self._move_turn_right_lat(pivot, drone, mag)
        target.longitude = self._move_turn_right_lon(pivot, drone, mag)
        target.altitude = self._move_up_down(drone.altitude, up_down)

        return target

    def _left_turn_mode(self, pivot, drone, up_down, mag):
        target = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0, RTC.GPS_FIX_NORMAL)

        target.latitude = self._move_turn_left_lat(pivot, drone, mag)
        target.longitude = self._move_turn_left_lon(pivot, drone, mag)
        target.altitude = self._move_up_down(drone.altitude, up_down)

        return target

    def _move_south_north(self, drone_lat, operation, speedparam):
        operation = self._analog_status_check(operation)

        diff = ((self._LatDiff10m / 10) *
                speedparam *
                self._LatLonCoefficient *
                operation)

        target_lat = drone_lat + diff

        if operation:
            target_lat = self._lat_range_check(target_lat)

        return target_lat

    def _move_east_west(self, drone_lon, operation, speedparam):
        operation = self._button_status_check(operation)

        diff = ((self._LonDiff10m / 10) *
                speedparam *
                self._LatLonCoefficient *
                operation)

        target_lon = drone_lon + diff

        if operation:
            target_lon = self._lon_range_check(target_lon)

        return target_lon

    def _move_up_down(self, drone_alt, operation):
        operation = self._button_status_check(operation)

        diff = self._AltCoefficient * operation
        target_alt = drone_alt + diff

        target_alt = self._alt_range_check(target_alt)

        return target_alt

    def _move_nose(self, drone, front_back, right_left, speedparam, mag):
        operation = 0.0

        front_back = self._analog_status_check(front_back)
        right_left = self._button_status_check(right_left)

        if (front_back == 1 and right_left == 1) or (front_back == -1 and right_left == -1):
            mag += 45.0
            operation = front_back
        elif (front_back == -1 and right_left == 1) or (front_back == 1 and right_left == -1):
            mag += 135.0
            operation = right_left
        elif front_back:
            operation = front_back
        elif right_left:
            mag += 90.0
            operation = right_left
            
        if mag > 360.0:
            mag = mag % 360.0
        elif mag < 0.0:
            mag = 360.0 - (abs(mag) % 360.0)


        latdiff = ((self._NoseLatDiff10m / 10) *
                pow(10, -5) *
                speedparam *
                self._LatLonCoefficient *
                math.cos(math.radians(mag)) *
                operation)
        londiff = (((self._NoseLonDiff10m / 10) *
                pow(10, -5) /
                math.cos(latdiff / (2 * math.pi))) *
                speedparam *
                self._LatLonCoefficient *
                math.sin(math.radians(mag)) *
                operation)

        target_lat = drone.latitude + latdiff
        target_lon = drone.longitude + londiff

        if operation:
            target_lat = self._lat_range_check(target_lat)             
            target_lon = self._lon_range_check(target_lon)

        return [target_lat, target_lon]

    #緯度、経度の１度の距離を求める
    #旋回計算の関数(_move_turn_right_lat,_move_turn_right_lon,_move_turn_left_lat,_move_turn_left_lon)の前に実行すること
    def _calc_lat_lon_degree_dist(self, lat):
        #緯度１度の距離（ｍ）
        self._lat_degree_dist = (EARTH_MERIDIAN_DIST*1000.0)/360.0
        #現在の緯度上における経度１度の距離（ｍ）
        self._lon_degree_dist = math.cos(math.pi * (lat / 180.0)) * ((EARTH_EQUATOR_DIST*1000.0)/360.0)
 
    #右旋回時の緯度を求める
    def _move_turn_right_lat(self, pivot, drone, mag):
        #度から距離(m)に変換
        lat_dist = (drone.latitude - pivot.latitude) * self._lat_degree_dist
        lon_dist = (drone.longitude - pivot.longitude) * self._lon_degree_dist
        #旋回計算
        target = ((lon_dist * math.sin(math.radians(-mag))) +
                 (lat_dist * math.cos(math.radians(-mag))))
        #距離(m)を度に変換
        target /= self._lat_degree_dist

        target += pivot.latitude

        return target

    def _move_turn_right_lon(self, pivot, drone, mag):
        #度から距離(m)に変換
        lat_dist = (drone.latitude - pivot.latitude) * self._lat_degree_dist
        lon_dist = (drone.longitude - pivot.longitude) * self._lon_degree_dist
        #旋回計算
        target = ((lon_dist * math.cos(math.radians(-mag))) -
                 (lat_dist * math.sin(math.radians(-mag))))
        #距離(m)を度に変換
        target /= self._lon_degree_dist

        target += pivot.longitude

        return target

    def _move_turn_left_lat(self, pivot, drone, mag):
        #度から距離(m)に変換
        lat_dist = (drone.latitude - pivot.latitude) * self._lat_degree_dist
        lon_dist = (drone.longitude - pivot.longitude) * self._lon_degree_dist
        #旋回計算
        target = ((lon_dist * math.sin(math.radians(mag))) +
                 (lat_dist * math.cos(math.radians(mag))))
        #距離(m)を度に変換
        target /= self._lat_degree_dist

        target += pivot.latitude

        return target

    def _move_turn_left_lon(self, pivot, drone, mag):
        #度から距離(m)に変換
        lat_dist = (drone.latitude - pivot.latitude) * self._lat_degree_dist
        lon_dist = (drone.longitude - pivot.longitude) * self._lon_degree_dist
        #旋回計算
        target = ((lon_dist * math.cos(math.radians(mag))) -
                 (lat_dist * math.sin(math.radians(mag))))
        #距離(m)を度に変換
        target /= self._lon_degree_dist

        target += pivot.longitude

        return target

    def get_target_cal(self, drone_target, leader_target, diff):
        drone_target.latitude = leader_target.latitude + diff.latitude
        drone_target.longitude = leader_target.longitude + diff.longitude
        if drone_target.altitude > 0.0:
            drone_target.altitude = leader_target.altitude + diff.altitude

    def _button_status_check(self, button):
        if button > 1:
            button = 1
        elif button < -1:
            button = -1
            
        return button

    def _analog_status_check(self, analog):
        if analog > 1:
            analog = 1
        elif analog < -1:
            analog = -1
            
        return analog

    def _alt_range_check(self, alt):
        if alt > self._MaxAlt:
            alt = self._MaxAlt
        elif alt < self._MinAlt:
            alt = self._MinAlt
            
        return alt

    def _lat_range_check(self, lat):
        if lat > 90.0:
            lat = 90.0
        elif lat < -90.0:
            lat = -90.0
        
        return lat

    def _lon_range_check(self, lon):
        if lon > 180.0:
            lonRem = lon % 360.0
            if lonRem > 180.0:
                lon = lonRem - 360.0
            else:
                lon = lonRem
        elif lon < -180.0:
            lonRem = abs(lon) % 360.0
            if lonRem > 180.0:
                lon = 360.0 - lonRem
            else:
                lon = -lonRem

        return lon
