#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file DroneIO.py
 @brief Wrapper of drone flight controller(ArduPilot)
 @date $Date$

 @author East Japan Accounting Center Co., Ltd

 (c)2017 The University of Aizu
 This software is released under the New BSD License.

"""
import sys
import time
import socket
import math
import csv
from datetime import datetime

# Import RTM module
import RTC
import OpenRTM_aist

import DroneServicePort_idl

# drone distance calc
import drone_distance_cal

# Setup logger
from logging import getLogger, StreamHandler
from logging import DEBUG, INFO

# Import DroneKit
from dronekit import connect, VehicleMode, LocationGlobalRelative, Command, Vehicle
from pymavlink import mavutil

sys.path.append(".")

logger = getLogger(__name__)
handler = StreamHandler()
logger.setLevel(DEBUG)
handler.setLevel(INFO)
logger.addHandler(handler)

# Constant declaration
TARGET_IP = "127.0.0.1:14550"  # Vehicle target address
MIN_MV_DIST = 2.0  # minimum move distance
HOME_LOC_GET_RETRY_CNT = 3  # The max num of downloads of the "home_location".
ARMED_RETRY_CNT = 2  # the max wait time of reflect "armed"
COMM_ERR_TIMER = 30  # communication error(w/ the drone controller
TAKEOFF_ARV_COEF = 0.99  # Takeoff arrival coefficient
TGT_ALT_ARV_COEF_MIN = 0.95  # Target alt arrival coefficient minimum
TGT_ALT_ARV_COEF_MAX = 1.05  # Target alt arrival coefficient maximum
HOVERING_ALLOW_RANGE = 0.2  # Hovering allowable range(unit:m)
VEHICLE_DEF_AIR_SPD = 3  # vehicle default speed(unit:m/s)
BATT_LIPO_VPC = 3.7  # Lipo Battery VPC(voltage per cell)
# Lipo Battery VPC(voltage per cell) percentage It means 3.7V = 50%
BATT_LIPO_VPC_PERCENT = 0.5
# Battery Parameters Tuning
BATT_MONI_V_ONLY = 3  # Analog Voltage Only
BATT_V_PIN_PIXHAWK = 2  # Battery Voltage sensing pin(Pixhawk)
BATT_C_PIN_PIXHAWK = -1  # Battery Current sensing pin(Disabled)
# RTL cause
RTL_CAUSE_BY_MANU = 1  # Manual operation
RTL_CAUSE_BY_ERR = 2  # Error
# manual RTL stage
RTL_STAGE_1ST = 1  # Goto RTL altitude
RTL_STAGE_2ND = 2  # Goto home-pos and keeping altitude
RTL_STAGE_3RD = 3  # Landing on the home-pos

# Error message
ERR_MSG_001 = "Socket error. Please re-startup this RTC."
ERR_MSG_002 = "Connection Timeout. Please re-startup this RTC."
ERR_MSG_003 = "Home_location timeout. Please re-startup this RTC."
ERR_MSG_004 = "Does not the home_location. Please re-startup this RTC."
ERR_MSG_005 = "Arm switch timeout. Please re-startup this RTC."
ERR_MSG_006 = "Communication with the GCS has been lost. "
ERR_MSG_007 = "Low battery detected. "
ERR_MSG_008 = "Start the RTL by error."
ERR_MSG_009 = "Robot ID is not registered!"

# Infomation message
INF_MSG_001 = "Connecting to vehicle on: (%(target_ip)s)"
INF_MSG_002 = "Waiting for home location..."
INF_MSG_003 = "Drone system: (%(sys_behavior)s)"
INF_MSG_004 = "Waiting for arming..."
INF_MSG_005 = "Switching to armed success."
INF_MSG_006 = "Start the takeoff."
INF_MSG_007 = "Takeoff complete."
INF_MSG_008 = "Arrived at the target alt."
INF_MSG_009 = "Start the RTL by manually."
INF_MSG_010 = "onExecute active->deactivate"
INF_MSG_011 = "RTC deactivated. Start the RTL."
INF_MSG_012 = "RTC deactivated. Switching to disarm."
INF_MSG_013 = "RTC deactivated. Will be continued RTL."
INF_MSG_014 = "Landed in the home position."
INF_MSG_015 = "ArduPilot F/W ver.: (%(ap_fw_ver)s)"

INF_MSG_100 = "Vehicle close."

# drone status declaration
STS_DISARM = 0
STS_ARMED = 1
STS_TAKEOFF = 2
STS_RTL = 3
STS_IN_MOVE = 4
STS_ALT_CHG = 5
STS_HOVERING = 6
STS_YAW_CHG = 7

# Copter parameter
#   Flight Mode
COPTER_PARAM_FLTMODE_STABILIZE = 0.0
COPTER_PARAM_FLTMODE_GUIDED = 4.0
COPTER_PARAM_FLTMODE_LOITER = 5.0

# drone status declaration(string)
droneio_sts_str = ("Disarm",
                   "Armed",
                   "Takeoff",
                   "RTL",
                   "In_Move",
                   "Alt_CHG",
                   "Hovering",
                   "Yaw_CHG"
                   )

# drone status Yaw_CHG tolerance
YAW_CHG_TOLERANCE = 1

# Turning direction
YAW_TURN_RIGHT = 1
YAW_TURN_LEFT = -1

# file mode
FILE_MODE_WRITE = "w"
FILE_MODE_ADD_WRITE = "a"

# Import Service implementation class
# <rtc-template block="service_impl">
from DroneServicePort_idl_example import *

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
droneio_spec = ["implementation_id", "DroneIO",
                "type_name",         "DroneIO",
                "description",       "Wrapper of drone flight controller(ArduPilot)",
                "version",           "1.4.1",
                "vendor",            "EAC",
                "category",          "Drone",
                "activity_type",     "STATIC",
                "max_instance",      "1",
                "language",          "Python",
                "lang_type",         "SCRIPT",
                "conf.default.BattCell", "4",
                "conf.default.RTLLowBatt", "30",
                "conf.default.BattChk", "on",
                "conf.default.TakeoffAlt", "10.0",
                "conf.default.RTLAlt", "1500",
                "conf.default.RTLAltFinal", "0",
                "conf.default.WpNavSpeed", "500",
                "conf.default.RTLLoitTime", "5000",
                "conf.default.WpNavSpeedDn", "100",
                "conf.default.LandSpeed", "50",
                "conf.default.DisarmDelay", "30",
                "conf.default.SystemBehavior", "single",
                "conf.default.RobotID", "-",
                "conf.__widget__.BattCell", "spin.1",
                "conf.__widget__.RTLLowBatt", "spin.1",
                "conf.__widget__.BattChk", "radio",
                "conf.__widget__.TakeoffAlt", "spin.0.1",
                "conf.__widget__.RTLAlt", "spin.1",
                "conf.__widget__.RTLAltFinal", "spin.1",
                "conf.__widget__.WpNavSpeed", "spin.50",
                "conf.__widget__.RTLLoitTime", "spin.5000",
                "conf.__widget__.WpNavSpeedDn", "spin.10",
                "conf.__widget__.LandSpeed", "spin.10",
                "conf.__widget__.DisarmDelay", "spin.1",
                "conf.__widget__.SystemBehavior", "radio",
                "conf.__widget__.RobotID", "text",
                "conf.__constraints__.BattCell", "2<=x<=7",
                "conf.__constraints__.RTLLowBatt", "20<=x<=60",
                "conf.__constraints__.BattChk", "(off,on)",
                "conf.__constraints__.TakeoffAlt", "x>=1.0",
                "conf.__constraints__.RTLAlt", "0<=x<=8000",
                "conf.__constraints__.RTLAltFinal", "-1<=x<=1000",
                "conf.__constraints__.WpNavSpeed", "0<=x<=2000",
                "conf.__constraints__.RTLLoitTime", "0<=x<=60000",
                "conf.__constraints__.WpNavSpeedDn", "0<=x<=500",
                "conf.__constraints__.LandSpeed", "30<=x<=200",
                "conf.__constraints__.DisarmDelay", "0<=x<=127",
                "conf.__constraints__.SystemBehavior", "(single,formation)",
                ""]
# </rtc-template>

##
# @class DroneIO
# @brief Wrapper of drone flight controller(ArduPilot)
#
# Overview:  Refer to the "機能仕様書_DroneIO_RTC.pdf" for details.
#
# In / Out:  Refer to the "機能仕様書_DroneIO_RTC.pdf" for details.
#
# Algorythm:  Refer to the "機能仕様書_DroneIO_RTC.pdf" for details.
#
#


class RTLAlt_i(Service__POA.RTLAlt):

    def set_rtl_alt(self, rtl_alt):
		self._rtl_alt = rtl_alt

    def get_rtl_alt(self):
        return self._rtl_alt


class DroneIO(OpenRTM_aist.DataFlowComponentBase):

    ##
    # @brief constructor
    # @param manager Maneger Object
    #
    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_inTarget = RTC.DroneTarget(RTC.Time(0, 0),
                                           0,
                                           RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                       100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                       0.0, 0, RTC.GPS_FIX_NORMAL),
                                           RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                           RTC.TimedShortSeq(RTC.Time(0, 0), [0, 0]))
        """
		Drone flight target position
		 - Semantics: - ID
		              - Target position(lat/lon/alt)
		              - Target speed(m/s)
		              - Target yaw
		              data[0] direction(1:right, -1:left, 0:none)
		              data[1] Absolute offset yaw(deg)
		"""
        self._inTargetIn = OpenRTM_aist.InPort("inTarget", self._d_inTarget)
        self._d_outGPSDrone = RTC.GPSData(RTC.Time(0, 0),
                                        RTC.GPSTime(0, 0),
                                        0.0, 0.0, 0.0,
                                        0.0, 0.0,
                                        0.0, 0.0,
                                        0.0,
                                        0,
                                        RTC.GPS_FIX_NORMAL)
        """
		GPS data for "Drone"
		 - Semantics: - Latitude
		              - Longitude
		              - Altitude(is relative to the "home_location")
		 - Unit: deg
		"""
        self._outGPSDroneOut = OpenRTM_aist.OutPort("outGPSDrone", self._d_outGPSDrone)
        self._d_outGyro = RTC.TimedOrientation3D(RTC.Time(0, 0),
                                                RTC.Orientation3D(0.0, 0.0, 0.0))
        """
		3 axis gyro sensor
		 - Semantics: - Roll angle
		              - Pitch angle
		              - Yaw angle
		 - Unit: rad
		"""
        self._outGyroOut = OpenRTM_aist.OutPort("outGyro", self._d_outGyro)
        self._d_outDroneCtrl = RTC.DroneCtrl(RTC.Time(0, 0),
                                             0,
                                             RTC.TimedString(RTC.Time(0, 0), ""),
                                             RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                         100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                         0.0, 0, RTC.GPS_FIX_NORMAL),
                                             RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                         100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                         0.0, 0, RTC.GPS_FIX_NORMAL),
                                             RTC.TimedFloatSeq(
                                                 RTC.Time(0, 0), [0.0, 0.0, 0.0]),
                                             RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                             RTC.TimedString(RTC.Time(0, 0), ""))
        """
		Drone flight control data
		 - Semantics: - GPS(Home)
		              - GPS(Drone)
		              - Acceleration
		              - Mag
		              - State
		"""
        self._outDroneCtrlOut = OpenRTM_aist.OutPort("outDroneCtrl", self._d_outDroneCtrl)
        self._d_outDroneMap = RTC.DroneMap(RTC.Time(0, 0),
                                           RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                       0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                       0.0, 0, RTC.GPS_FIX_NORMAL),
                                           RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                       0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                       0.0, 0, RTC.GPS_FIX_NORMAL),
                                           RTC.TimedFloat(RTC.Time(0, 0), 0),
                                           RTC.TimedString(RTC.Time(0, 0), 0))
        """
		Drone map info
		 - Semantics: - GPS(Drone)
		              - Target Position
		              - Mag
		              - State
		"""
        self._outDroneMapOut = OpenRTM_aist.OutPort("outDroneMap", self._d_outDroneMap)
        self._d_outDroneDB = RTC.DroneDB(RTC.Time(0, 0),
                                         RTC.TimedString(RTC.Time(0, 0), ""),
                                         RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                     100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                     0.0, 0, RTC.GPS_FIX_NORMAL),
                                         RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                     100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                     0.0, 0, RTC.GPS_FIX_NORMAL),
                                         RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                         RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                         RTC.TimedULong(RTC.Time(0, 0), 0),
                                         RTC.TimedOrientation3D(
                                             RTC.Time(0, 0), RTC.Orientation3D(0.0, 0.0, 0.0)),
                                         RTC.TimedFloatSeq(
                                             RTC.Time(0, 0), [0.0, 0.0, 0.0]),
                                         RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                         RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                         RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                         RTC.TimedString(RTC.Time(0, 0), ""),
                                         RTC.TimedString(RTC.Time(0, 0), ""),
                                         RTC.TimedString(RTC.Time(0, 0), ""),
                                         RTC.TimedLongSeq(RTC.Time(0, 0), [
                                                          0, 0, 0, 0, 0, 0, 0, 0, 0]),
                                         RTC.TimedFloatSeq(RTC.Time(0, 0), [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]))
        """
		Write sensor info to DB
		"""
        self._outDroneDBOut = OpenRTM_aist.OutPort("outDroneDB", self._d_outDroneDB)
        self._d_outDroneView = RTC.DroneView(RTC.Time(0, 0),
                                            RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                        100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                        0.0, 0, RTC.GPS_FIX_NORMAL),
                                            RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                        100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                        0.0, 0, RTC.GPS_FIX_NORMAL),
                                            RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                            RTC.TimedFloatSeq(RTC.Time(0, 0), [0.0, 0.0]),
                                            RTC.TimedULong(RTC.Time(0, 0), 0),
                                            RTC.TimedOrientation3D(
                                                RTC.Time(0, 0), RTC.Orientation3D(0.0, 0.0, 0.0)),
                                            RTC.TimedFloatSeq(
                                                RTC.Time(0, 0), [0.0, 0.0, 0.0]),
                                            RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                            RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                            RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                            RTC.TimedString(RTC.Time(0, 0), ""),
                                            RTC.TimedString(RTC.Time(0, 0), ""),
                                            RTC.TimedString(RTC.Time(0, 0), ""))
        """
		Write sensor info to Viewer
		 - Unit: For detail, see the "機能仕様書_DroneIO"
		"""
        self._outDroneViewOut=OpenRTM_aist.OutPort("outDroneView", self._d_outDroneView)

        """
        Provide the drone info
        """
        self._ProvDroneInfoPort = OpenRTM_aist.CorbaPort("ProvDroneInfo")

        """
        Provide the  RTL altitude
        """
        self._setRTLAlt = RTLAlt_i()

        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">
        """
		The number of cells of the battery mounted on the drone.
		 - Name: BattCell BattCell
		 - DefaultValue: 4
		 - Unit: cell
		 - Range: 2<=x<=7
		"""
        self._BattCell = [4]
        """
		RTL is done when the battery level falls below the threshold.
		 - Name: RTLLowBatt RTLLowBatt
		 - DefaultValue: 30
		 - Unit: %
		 - Range: 20<=x<=60
		"""
        self._RTLLowBatt = [30]
        """
		Whether the battery check
		 - Name: BattChk BattChk
		 - DefaultValue: on
		 - Range: (off, on)
		"""
        self._BattChk = ['on']
        """
		Takeoff  altitude
		 - Name: TakeoffAlt TakeoffAlt
		 - DefaultValue: 10.0
		 - Unit: m
		 - Range: 1.0<=x
		"""
        self._TakeoffAlt = [10.0]
        """
		The minimum relative altitude the model will move to before RTL. Set to
		zero to return at current altitude.
		 - Name: RTLAlt RTLAlt
		 - DefaultValue: 1500
		 - Unit: cm
		 - Range: 0<=x<=8000
		"""
        self._RTLAlt = [1500]
        """
		This is the altitude the vehicle will move to as the final stage of RTL or
		after completing a mission. Set to zero to land.
		 - Name: RTLAltFinal RTLAltFinal
		 - DefaultValue: 0
		 - Unit: cm
		 - Range: -1<=x<=1000
		"""
        self._RTLAltFinal = [0]
        """
		Waypoint horizontal speed target.
		(This parameter is an alternative to "RTL_SPEED".)
		 - Name: WpNavSpeed WpNavSpeed
		 - DefaultValue: 500
		 - Unit: cm/s
		 - Range: 0<=x<=2000
		"""
        self._WpNavSpeed = [500]
        """
		Time  to loiter above home before beginning final descent.
		 - Name: RTLLoitTime RTLLoitTime
		 - DefaultValue: 5000
		 - Unit: msec
		 - Range: 0<=x<=60000
		"""
        self._RTLLoitTime = [5000]
        """
		Waypoint descent speed target.
		(This parameter is an alternative to "LAND_SPEED_HIGH".)
		 - Name: WpNavSpeedDn WpNavSpeedDn
		 - DefaultValue: 100
		 - Unit: cm/s
		 - Range: 0<=x<=500
		"""
        self._WpNavSpeedDn = [100]
        """
		The descent speed for the final stage of landing in cm/s.
		 - Name: LandSpeed LandSpeed
		 - DefaultValue: 50
		 - Unit: cm/s
		 - Range: 30<=x<=200
		"""
        self._LandSpeed = [50]
        """
		Delay before automatic disarm in seconds. A value of zero disables auto
		disarm.
		 - Name: DisarmDelay DisarmDelay
		 - DefaultValue: 30
		 - Unit: sec
		 - Range: 0<=x<=127
		"""
        self._DisarmDelay = [30]
        """
		This RTC has different behaviors depending on the system, so select which
		system to behave.
		 - Name:  SystemBehavior
		 - DefaultValue: single
		"""
        self._SystemBehavior = ['single']
        """
		 - Name:  RobotID
		 - DefaultValue: -
		"""
        self._RobotID = ['-']

        # </rtc-template>

        self._vehicle = None  # vehicle object
        self._flight_time_start = 0  # Flight Start Time
        self._drone_sts = STS_DISARM
        self._comm_err_timer = 0  # communication error check timer
        self._takeoff_alt = 0
        self._tgt_alt = 0
        self._rtl_cause = 0
        self._rtl_stage = 0

        self._file_name = ""
        self._file = None
        self._writer = None
        self._header = ["Category", "ActStartDate","ActStartTime", "RcvDate", "RcvTime", "RcvRobotID", "RcvID",
                       "TargetLat", "TargetLon", "TargetAlt", "TargetSpd", "TargetYawDir", "TargetYawDeg",
                       "SndDate", "SndTime", "SndRobotID", "SndID",
                       "HomeLat", "HomeLon", "HomeAlt", "RobotLat", "RobotLon", "RobotAlt",
                       "AccX", "AccY", "AccZ", "Mag", "State",
                       "FltMode", "RTCState", "Err", "BattV", "GyroR", "GyroP", "GyroY", "Temp", "Pres"
                       ]
        self._act_start_date = ""
        self._act_start_time = ""

        # read-info
        self._gps_home = RTC.GPSData(RTC.Time(0, 0),
                                        RTC.GPSTime(0, 0),
                                        100.0, 200.0, 0.0,
                                        0.0, 0.0,
                                        0.0, 0.0,
                                        0.0,
                                        0,
                                        RTC.GPS_FIX_NORMAL)
        self._gps_drone = RTC.GPSData(RTC.Time(0, 0),
                                        RTC.GPSTime(0, 0),
                                        100.0, 200.0, 0.0,
                                        0.0, 0.0,
                                        0.0, 0.0,
                                        0.0,
                                        0,
                                        RTC.GPS_FIX_NORMAL)
        self._batt = RTC.TimedFloatSeq(RTC.Time(0, 0), [0.0, 0.0])
        self._gyro = RTC.TimedOrientation3D(RTC.Time(0, 0), RTC.Orientation3D(0.0, 0.0, 0.0))
        self._accel = RTC.TimedFloatSeq(RTC.Time(0, 0), [0.0, 0.0, 0.0])
        self._mag = RTC.TimedFloat(RTC.Time(0, 0), 0.0)
        self._temp = RTC.TimedFloat(RTC.Time(0, 0), 0.0)
        self._pres = RTC.TimedFloat(RTC.Time(0, 0), 0.0)

        self._fltTime = RTC.TimedULong(RTC.Time(0, 0), 0)
        self._fltMode = RTC.TimedString(RTC.Time(0, 0), "")
        self._state = RTC.TimedString(RTC.Time(0, 0), "")
        self._sysErr = RTC.TimedString(RTC.Time(0, 0), "")

        # set rtl alt to the service port
        self._setRTLAlt.set_rtl_alt(self._RTLAlt[0] / 100)

    ##
    # RTC Initialize
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry()
    #
    # @return RTC::ReturnCode_t
    #
    #
    def onInitialize(self):
        # Bind variables and configuration variable
        self.bindParameter("BattCell", self._BattCell, "4")
        self.bindParameter("RTLLowBatt", self._RTLLowBatt, "30")
        self.bindParameter("BattChk", self._BattChk, "on")
        self.bindParameter("TakeoffAlt", self._TakeoffAlt, "10.0")
        self.bindParameter("RTLAlt", self._RTLAlt, "1500")
        self.bindParameter("RTLAltFinal", self._RTLAltFinal, "0")
        self.bindParameter("WpNavSpeed", self._WpNavSpeed, "500")
        self.bindParameter("RTLLoitTime", self._RTLLoitTime, "5000")
        self.bindParameter("WpNavSpeedDn", self._WpNavSpeedDn, "100")
        self.bindParameter("LandSpeed", self._LandSpeed, "50")
        self.bindParameter("DisarmDelay", self._DisarmDelay, "30")
        self.bindParameter("SystemBehavior", self._SystemBehavior, "single")
        self.bindParameter("RobotID", self._RobotID, "-")

        # Set InPort buffers
        self.addInPort("inTarget", self._inTargetIn)

        # Set OutPort buffers
        self.addOutPort("outGPSDrone", self._outGPSDroneOut)
        self.addOutPort("outGyro", self._outGyroOut)
        self.addOutPort("outDroneCtrl", self._outDroneCtrlOut)
        self.addOutPort("outDroneMap", self._outDroneMapOut)
        self.addOutPort("outDroneDB", self._outDroneDBOut)
        self.addOutPort("outDroneView", self._outDroneViewOut)

        # Set service provider to Ports
        self._ProvDroneInfoPort.registerProvider("setRTLAlt", "Service::RTLAlt", self._setRTLAlt)

        # Set service consumers to Ports

        # Set CORBA Service Ports
        self.addPort(self._ProvDroneInfoPort)

        return RTC.RTC_OK

        ##
        # Close vehicle object
        #
        # The finalize action (on ALIVE->END transition)
        # formaer rtc_exiting_entry()
        #
        # @return RTC::ReturnCode_t

        #
    def onFinalize(self):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()

        self._vehicle.close()
        logger.info(INF_MSG_100)

        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Exit")
        self._close_rtc_viewer_log()

        return RTC.RTC_OK

        ##
        # - Connecting to the vehicle(default target adress: 127.0.0.1:14550).
    #	- Download the "home_location"
        #
        # The startup action when ExecutionContext startup
        # former rtc_starting_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onStartup(self, ec_id):

        rtn = False

        if self._is_vehicle_connected() is True:
            if self._is_home_location_acquired() is True:
                rtn = True

        if rtn is True:
            return RTC.RTC_OK
        else:
            return RTC.RTC_ERROR

    #	##
    #	#
    #	# The shutdown action when ExecutionContext stop
    #	# former rtc_stopping_entry()
    #	#
    #	# @param ec_id target ExecutionContext Id
    #	#
    #	# @return RTC::ReturnCode_t
    #	#
    #	#
    # def onShutdown(self, ec_id):
    #
    #	return RTC.RTC_OK

        ##
        # - Switch to armed and wait for reflection.
    #	- Vehicle mode is set to GUIDED
    #	- Parameter setting
    #	- Start the counting of the flight timer.
        #
        # The activated action (Active state entry action)
        # former rtc_active_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        # @pre "home_location" exists
        #
    def onActivated(self, ec_id):

        rtn = False

        self._file_name = ""
        self._file = None
        self._writer = None
        self._act_start_date, self._act_start_time = self._get_datetime_to_str()

        self._init_read_info()

        logger.info(INF_MSG_003 % {"sys_behavior": self._SystemBehavior[0]})

        # low-battery threshold calculation
        self._batt_thr = (BATT_LIPO_VPC - (BATT_LIPO_VPC_PERCENT -
                                           (self._RTLLowBatt[0] / 100.0))) * self._BattCell[0]

        # robot ID regist check
        if self._RobotID[0] != '-':
            # battery check
            if self._is_batt_state() is True:
                # RTL or not
                if self._drone_sts != STS_RTL:
                    # home_location check
                    if self._vehicle.home_location is not None:
                        # Change the "GUIDED" mode
                        self._vehicle.mode = VehicleMode("GUIDED")

                        if self._switching_to_armed() is True:
                            # Battery parameters tuning
                            self._batt_param_tuning()
                            # Takeoff paramter tuning
                            self._vehicle.parameters[
                                'DISARM_DELAY'] = self._DisarmDelay[0]
                            # RTL paramter tuning
                            self._rtl_param_tuning()
                            # Change the arm status
                            self._drone_sts = STS_ARMED
                            self._write_io_state()
                            # Error Info clear
                            self._write_io_err_inf("")
                            self._init_val_rtl()
                            # Flight time count to start
                            self._flight_time_start = time.time()
                            rtn = True
                        else:
                            self._write_io_err_inf(ERR_MSG_005)
                    else:
                        self._write_io_err_inf(ERR_MSG_004)
                else:
                    # in RTL
                    rtn = True
            else:
                self._write_io_err_inf(ERR_MSG_007)
        else:
            self._write_io_err_inf(ERR_MSG_009)

        # viewer-log open and write(header-row)
        self._fltMode.data = self._vehicle.mode.name
        self._log_row = ["NULL"] * len(self._header)
        self._ready_to_log_write(FILE_MODE_WRITE)

        if rtn is True:
            return RTC.RTC_OK
        else:
            return RTC.RTC_ERROR

        ##
        # RTL if in flight, otherwise switch to disarm if RTL is done.
        #
        # The deactivated action (Active state exit action)
        # former rtc_active_exit()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onDeactivated(self, ec_id):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()

        if self._drone_sts != STS_RTL:
            if self._drone_sts not in [STS_DISARM, STS_ARMED]:
                self._rtl_by_err()
                self._write_flt_mode()
                logger.info(INF_MSG_011)
            else:
                self._vehicle.armed = False
                self._drone_sts = STS_DISARM
                logger.info(INF_MSG_012)
            self._write_io_state()
        else:
            if self._vehicle.armed is True:
                if self._SystemBehavior[0] == 'formation':
                    if self._vehicle.mode.name != "RTL":
                        self._rtl_by_err()
                        self._write_flt_mode()
                        logger.info(INF_MSG_011)
                    else:
                        logger.info(INF_MSG_013)
                else:
                    logger.info(INF_MSG_013)
            else:
                self._drone_sts = STS_DISARM
                self._write_io_state()
                logger.info(INF_MSG_014)

        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Inactive")
        self._close_rtc_viewer_log()

        return RTC.RTC_OK

        ##
        # - Get and output of the sensor info through the DroneKit-Python API.
    #	- Flight instruction through the DronKit-Python API.
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onExecute(self, ec_id):

        rtn = True

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()
        self._d_inTarget = RTC.DroneTarget(RTC.Time(0, 0),
                                           0,
                                           RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(0, 0),
                                                       100.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                       0.0, 0, RTC.GPS_FIX_NORMAL),
                                           RTC.TimedFloat(RTC.Time(0, 0), 0.0),
                                           RTC.TimedShortSeq(RTC.Time(0, 0), [0, 0]))

        # Armed Check
        if self._vehicle.armed is True:
            # RTL or not
            if self._drone_sts != STS_RTL:
                # battery check
                if self._is_batt_state() is True:
                    # Read the target position
                    if self._vehicle.mode.name == "GUIDED":
                        if self._is_read_info() is True:
                            self._comm_err_timer = 0
                            self._flight_ctrl_core()
                        else:
                            self._comm_err_timer += 1
                            if self._is_gcs_comm_err() is True:
                                # GCS communication error
                                rtn = False

                    # judge the target position arrive
                    self._arv_judge()
                else:
                    # Low batt detected. Start the RTL.
                    self._set_io_err_inf(ERR_MSG_007)
                    rtn = False
            else:
                # RTL for formation-flight
                if self._SystemBehavior[0] == 'formation':
                    if self._rtl_cause == RTL_CAUSE_BY_MANU:
                        self._continuous_manu_rtl()

            # Read & output the sensor info
            self._read_info()
            self._set_info("Active")
            self._write_info()
        else:
            # switching from armed to disarm
            self.deactivate(ec_id)
            logger.info(INF_MSG_010)

        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Active")

        if rtn is True:
            return RTC.RTC_OK
        else:
            return RTC.RTC_ERROR

    #	##
    #	#
    #	# The aborting action when main logic error occurred.
    #	# former rtc_aborting_entry()
    #	#
    #	# @param ec_id target ExecutionContext Id
    #	#
    #	# @return RTC::ReturnCode_t
    #	#
    #	#
    # def onAborting(self, ec_id):
    #
    #	return RTC.RTC_OK

        ##
        # RTL if in flight
        #
        # The error action in ERROR state
        # former rtc_error_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onError(self, ec_id):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()

        # Start RTL, if in flight
        if self._drone_sts not in [STS_DISARM, STS_ARMED, STS_RTL]:
            self._rtl_by_err()

        # Read & output the sensor info
        self._read_info()
        self._set_info("Error")
        self._write_info()

        # State change to disarm, if landed
        if self._vehicle.armed is False:
            self._drone_sts = STS_DISARM
            self._write_io_state()
            logger.info(INF_MSG_014)
            self.reset(ec_id)

        if self._file is None:
            self._ready_to_log_write(FILE_MODE_ADD_WRITE)
        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Error")

        return RTC.RTC_OK

        ##
        #
        # The reset action that is invoked resetting
        # This is same but different the former rtc_init_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onReset(self, ec_id):

        self._act_start_date, self._act_start_time = self._get_datetime_to_str()
        self._log_row = ["NULL"] * len(self._header)
        self._write_rtc_viewer_log("Inactive")
        self._close_rtc_viewer_log()

        return RTC.RTC_OK

    #	##
    #	#
    #	# The state update action that is invoked after onExecute() action
    #	# no corresponding operation exists in OpenRTm-aist-0.2.0
    #	#
    #	# @param ec_id target ExecutionContext Id
    #	#
    #	# @return RTC::ReturnCode_t
    #	#

    #	#
    # def onStateUpdate(self, ec_id):
    #
    #	return RTC.RTC_OK

    #	##
    #	#
    #	# The action that is invoked when execution context's rate is changed
    #	# no corresponding operation exists in OpenRTm-aist-0.2.0
    #	#
    #	# @param ec_id target ExecutionContext Id
    #	#
    #	# @return RTC::ReturnCode_t
    #	#
    #	#
    # def onRateChanged(self, ec_id):
    #
    #	return RTC.RTC_OK

    def _is_vehicle_connected(self):
    
        rtn = False

        try:
            logger.info(INF_MSG_001 % {"target_ip": TARGET_IP})
            self._vehicle = connect(
                TARGET_IP, wait_ready=True, vehicle_class=MyVehicle)
            self._vehicle.wait_ready('autopilot_version')
            logger.info(INF_MSG_015 % {"ap_fw_ver": self._vehicle.version})
            rtn = True
        except socket.error:
            logger.error(ERR_MSG_001)
        except dronekit.APIException:
            logger.error(ERR_MSG_002)

        return rtn

    def _is_home_location_acquired(self):

        rtn = False

        for retrycnt in range(HOME_LOC_GET_RETRY_CNT):
            cmds = self._vehicle.commands
            cmds.download()
            cmds.wait_ready()
            if not self._vehicle.home_location:
                logger.info(INF_MSG_002)
                time.sleep(1)
            else:
                logger.info("%s", self._vehicle.home_location)
                rtn = True
                break

        if rtn is False:
            logger.error(ERR_MSG_003)

        return rtn

    def _switching_to_armed(self):

        rtn = False

        self._vehicle.armed = True
        for retrycnt in range(ARMED_RETRY_CNT):
            if self._vehicle.armed is not True:
                logger.info(INF_MSG_004)
                time.sleep(1)
            else:
                logger.info(INF_MSG_005)
                rtn = True
                break

        return rtn

    def _batt_param_tuning(self):

        self._vehicle.parameters['BATT_MONITOR'] = BATT_MONI_V_ONLY
        self._vehicle.parameters['BATT_VOLT_PIN'] = BATT_V_PIN_PIXHAWK
        self._vehicle.parameters['BATT_CURR_PIN'] = BATT_C_PIN_PIXHAWK

    def _rtl_param_tuning(self):

        self._vehicle.parameters['RTL_ALT'] = self._RTLAlt[0]
        self._vehicle.parameters['RTL_ALT_FINAL'] = self._RTLAltFinal[0]
        self._vehicle.parameters['WPNAV_SPEED'] = self._WpNavSpeed[0]
        self._vehicle.parameters['WPNAV_SPEED_DN'] = self._WpNavSpeedDn[0]
        self._vehicle.parameters['RTL_LOIT_TIME'] = self._RTLLoitTime[0]
        self._vehicle.parameters['LAND_SPEED'] = self._LandSpeed[0]

    def _get_propo_fltmode(self):

        fltmd_val = COPTER_PARAM_FLTMODE_GUIDED

        if 0 <= self._vehicle.channels['5'] <= 1230:
            fltmd_val = self._vehicle.parameters['FLTMODE1']
        elif 1231 <= self._vehicle.channels['5'] <= 1749:
            fltmd_val = self._vehicle.parameters['FLTMODE2']
        elif 1750 <= self._vehicle.channels['5']:
            fltmd_val = self._vehicle.parameters['FLTMODE6']
        else:
            pass

        return fltmd_val

    def _is_read_info(self):

        rtn = False

        # Data port
        if self._inTargetIn.isNew():
            self._d_inTarget = self._inTargetIn.read()
            OpenRTM_aist.setTimestamp(self._d_inTarget)

            read_pos = self._is_read_target_pos()
            read_spd = self._is_read_target_spd()
            read_yaw = self._is_read_target_yaw()
            logger.debug("RcvID:    %d", self._d_inTarget.id)
            logger.debug("pos:      %f %f %f",
                         self._d_inTarget.pos.latitude,
                         self._d_inTarget.pos.longitude,
                         self._d_inTarget.pos.altitude)
            logger.debug("speed:    %f", self._d_inTarget.speed.data)
            logger.debug("yaw:      %s", self._d_inTarget.yaw.data)

            # Read-info initialize
            self._init_read_info()
            
            # Read from flight controller
            self._read_gps_home()
            self._read_gps_drone()
            self._read_mag()

            if (read_pos and read_spd and read_yaw):
                rtn = True

        return rtn

    def _is_read_target_pos(self):

        rtn = False
        valid_lat = False
        valid_lon = False
        valid_alt = False

        valid_lat = self._is_valid_lat(self._d_inTarget.pos.latitude)
        valid_lon = self._is_valid_lon(self._d_inTarget.pos.longitude)
        valid_alt = self._is_valid_alt(self._d_inTarget.pos.altitude)

        if valid_lat and valid_lon and valid_alt:
            rtn = True

        return rtn

    def _is_read_target_spd(self):

        rtn = False

        if self._d_inTarget.speed.data != 0:
            self._vehicle.airspeed = self._d_inTarget.speed.data
        else:
            self._vehicle.airspeed = VEHICLE_DEF_AIR_SPD
        rtn = True

        return rtn

    def _is_read_target_yaw(self):

        rtn = False

        valid_yaw = self._is_valid_yaw_deg(self._d_inTarget.yaw.data[1])
        if valid_yaw:
            rtn = True

        return rtn

    def _flight_ctrl_core(self):

        if self._drone_sts == STS_ARMED:
            # Start takeoff, if takeoff request.
            self._takeoff()
        elif self._drone_sts in [STS_IN_MOVE, STS_ALT_CHG, STS_HOVERING, STS_YAW_CHG]:
            if self._is_rtl_req() is True:
                # RTL by manually
                self._manual_rtl()
            else:
                # Move to the target position and yaw.
                self._move_to_target_pos()
                self._move_to_target_yaw()
        else:
            # None, if the state is RTL or take off
            pass

    def _is_gcs_comm_err(self):

        rtn = False

        if self._comm_err_timer == COMM_ERR_TIMER:
            self._comm_err_timer = 0
            self._set_io_err_inf(ERR_MSG_006)
            rtn = True

        return rtn

    def _takeoff(self):
        if self._d_inTarget.pos.altitude == self._TakeoffAlt[0]:
            self._takeoff_alt = self._d_inTarget.pos.altitude
            self._vehicle.simple_takeoff(self._takeoff_alt)
            self._drone_sts = STS_TAKEOFF
            logger.info(INF_MSG_006)
        else:
            logger.error("[E]Invalid takeoff alt(%f[m]) received!!",
                         self._d_inTarget.pos.altitude)

    def _is_rtl_req(self):

        rtn = False

        if self._d_inTarget.pos.latitude == self._gps_home.latitude:
            if self._d_inTarget.pos.longitude == self._gps_home.longitude:
                if self._d_inTarget.pos.altitude == self._gps_home.altitude:
                    rtn = True

        return rtn

    def _move_to_target_pos(self):

        dist_cal = drone_distance_cal.DroneMathmetics()
        dist = dist_cal.distance_math(self._gps_drone.longitude,
                                      self._d_inTarget.pos.longitude,
                                      self._gps_drone.latitude,
                                      self._d_inTarget.pos.latitude,
                                      0,
                                      0)

        logger.debug("dist:%f[m]", dist)

        # "MIN_MV_DIST" means mininum move distance
        # See "機能仕様書_DroneController" for details
        if dist <= MIN_MV_DIST:
            # altitude change, if change value is allowable range
            self._alt_chg_req_alt = self._d_inTarget.pos.altitude
            if round(abs(self._d_inTarget.pos.altitude - self._gps_drone.altitude), 7) > HOVERING_ALLOW_RANGE:
                self._drone_sts = STS_ALT_CHG
            else:
                self._drone_sts = STS_HOVERING
        else:
            self._move_target_lat = self._d_inTarget.pos.latitude
            self._move_target_lon = self._d_inTarget.pos.longitude
            self._drone_sts = STS_IN_MOVE

        # Move to the target point
        point = LocationGlobalRelative(
            self._d_inTarget.pos.latitude,
            self._d_inTarget.pos.longitude,
            self._d_inTarget.pos.altitude)
        self._vehicle.simple_goto(point)

    def _move_to_target_yaw(self):
        if self._d_inTarget.yaw.data[0] in [YAW_TURN_RIGHT, YAW_TURN_LEFT]:
            self._condition_yaw()

    def _condition_yaw(self):

        # create the CONDITION_YAW command using command_long_encode()
        msg = self._vehicle.message_factory.command_long_encode(
            0, 0,  # target system, target component
            mavutil.mavlink.MAV_CMD_CONDITION_YAW,  # command
            0,  # confirmation
            self._d_inTarget.yaw.data[1],   # param 1, yaw in degrees
            0,                              # param 2, yaw speed deg/s
            self._d_inTarget.yaw.data[0],   # param 3, direction -1 ccw, 1 cw
            0,                              # param 4, abusolute direction
            0, 0, 0)                        # param 5 ~ 7 not used

        # send command to vehicle
        self._vehicle.send_mavlink(msg)

        # state yaw change
        diff = abs(self._d_inTarget.yaw.data[1] - self._mag.data)
        if diff > 180:
            diff = 360 - diff

        if diff > YAW_CHG_TOLERANCE:
            self._drone_sts = STS_YAW_CHG

    def _arv_judge(self):

        if self._drone_sts == STS_TAKEOFF:
            if self._takeoff_alt * TAKEOFF_ARV_COEF <= self._gps_drone.altitude:
                logger.info(INF_MSG_007)
                self._drone_sts = STS_HOVERING

    def _init_val_rtl(self):

        self._rtl_cause = 0
        self._rtl_stage = 0

    def _manual_rtl(self):
        self._rtl_cause = RTL_CAUSE_BY_MANU
        if self._SystemBehavior[0] == 'formation':
            # Go to RTLAlt
            self._drone_sts = STS_RTL
            self._rtl_stage = RTL_STAGE_1ST
            self._tgt_alt = self._RTLAlt[0] / 100
            point = LocationGlobalRelative(
                self._gps_drone.latitude,
                self._gps_drone.longitude,
                self._tgt_alt)
            self._vehicle.simple_goto(point)
        else:
            self._start_rtl()

    def _start_rtl(self):

        self._vehicle.mode = VehicleMode("RTL")
        self._drone_sts = STS_RTL
        if self._rtl_cause == RTL_CAUSE_BY_MANU:
            logger.info(INF_MSG_009)
        elif self._rtl_cause == RTL_CAUSE_BY_ERR:
            self._set_io_err_inf(ERR_MSG_008)

    def _continuous_manu_rtl(self):

        if self._rtl_stage == RTL_STAGE_1ST:
            if self._is_arv_tgt_alt() is True:
                self._rtl_stage = RTL_STAGE_2ND
                point = LocationGlobalRelative(
                    self._gps_home.latitude,
                    self._gps_home.longitude,
                    self._tgt_alt)
                self._vehicle.simple_goto(point)

        elif self._rtl_stage == RTL_STAGE_2ND:
            read_tgt = self._is_read_info()
            if read_tgt:
                if self._is_rtl_req():
                    self._rtl_stage = RTL_STAGE_3RD
                    self._start_rtl()

        else:
            pass

    def _rtl_by_err(self):

        self._rtl_cause = RTL_CAUSE_BY_ERR
        self._start_rtl()

    def _is_arv_tgt_alt(self):

        rtn = False

        tgt_pos_min = self._tgt_alt * TGT_ALT_ARV_COEF_MIN
        tgt_pos_max = self._tgt_alt * TGT_ALT_ARV_COEF_MAX
        if tgt_pos_min <= self._gps_drone.altitude <= tgt_pos_max:
            logger.info(INF_MSG_008)
            rtn = True

        return rtn

    def _init_read_info(self):

        self._is_read_gps_home = False
        self._is_read_gps_drone = False
        self._is_read_battV = False
        self._is_read_battA = False
        self._is_read_gyro = False
        self._is_read_accel = False
        self._is_read_mag = False
        self._is_read_temp = False
        self._is_read_pres = False

        self._gps_home.latitude = 100.0
        self._gps_home.longitude = 200.0
        self._gps_home.altitude = 0.0

        self._gps_drone.latitude = 100.0
        self._gps_drone.longitude = 200.0
        self._gps_drone.altitude = 0.0

        self._batt.data = [0.0, 0.0]
        self._gyro.data = RTC.Orientation3D(0.0, 0.0, 0.0)
        self._accel.data = [0.0, 0.0, 0.0]
        self._mag.data = 0.0
        self._temp.data = 0.0
        self._pres.data = 0.0

    def _read_info(self):

        self._read_gps_home()
        self._read_gps_drone()
        self._read_batt()
        self._read_gyro()
        self._read_accel()
        self._read_mag()
        self._read_temperature()
        self._read_pressure()

        self._read_flt_time()
        self._read_flt_mode()
        self._read_io_state()

    def _set_info(self, rtc_state):
    
        self._set_drone_ctrl()
        self._set_map(rtc_state)
        self._set_view()
        self._set_db()

    def _write_info(self):

        self._write_drone_ctrl()
        self._write_map()
        self._write_view()
        self._write_db()
        self._write_gps_drone()
        self._write_gyro()

    def _read_gps_home(self):

        self._is_read_gps_home = False
        if self._vehicle.home_location is not None:
            tmp_home_location = self._vehicle.home_location
            self._gps_home.tm = RTC.Time(0, 0)
            self._gps_home.timeFromGPS = RTC.GPSTime(0, 0)
            self._gps_home.latitude = round(tmp_home_location.lat, 10)
            self._gps_home.longitude = round(tmp_home_location.lon, 10)
            self._gps_home.altitude = round(tmp_home_location.alt, 2)
            self._gps_home.horizontalError = 0.0
            self._gps_home.verticalError = 0.0
            self._gps_home.heading = 0.0
            self._gps_home.horizontalSpeed = 0.0
            self._gps_home.verticalSpeed = 0.0
            self._gps_home.numSatellites = 0
            self._gps_home.fixType = RTC.GPS_FIX_NORMAL
            self._is_read_gps_home = True

    def _read_gps_drone(self):

        self._is_read_gps_drone = False
        if self._vehicle.location.global_relative_frame.lat is not None:
            if self._vehicle.location.global_relative_frame.lon is not None:
                if self._vehicle.location.global_relative_frame.alt is not None:
                    self._gps_drone.latitude = round(self._vehicle.location.global_relative_frame.lat, 10)
                    self._gps_drone.longitude = round(self._vehicle.location.global_relative_frame.lon, 10)
                    self._gps_drone.altitude = round(self._vehicle.location.global_relative_frame.alt, 2)
                    self._gps_drone.tm = RTC.Time(0, 0)
                    self._gps_drone.timeFromGPS = RTC.GPSTime(0, 0)
                    self._gps_drone.horizontalError = 0.0
                    self._gps_drone.verticalError = 0.0
                    self._gps_drone.heading = 0.0
                    self._gps_drone.horizontalSpeed = 0.0
                    self._gps_drone.verticalSpeed = 0.0
                    self._gps_drone.numSatellites = 0
                    self._gps_drone.fixType = RTC.GPS_FIX_NORMAL
                    self._is_read_gps_drone = True

    def _write_gps_drone(self):

        if self._is_read_gps_drone:
            self._d_outGPSDrone.latitude = self._gps_drone.latitude
            self._d_outGPSDrone.longitude = self._gps_drone.longitude
            self._d_outGPSDrone.altitude = self._gps_drone.altitude
            self._outGPSDroneOut.write()

    def _read_batt(self):

        # Voltage
        self._is_read_battV = False
        if self._vehicle.battery.voltage is not None:
            self._batt.tm = RTC.Time(0, 0)
            self._batt.data[0] = round(self._vehicle.battery.voltage, 2)
            self._is_read_battV = True

        # Current
        self._is_read_battA = False
        if self._vehicle.battery.current is not None:
            self._batt.tm = RTC.Time(0, 0)
            self._batt.data[1] = round(self._vehicle.battery.current / 1000, 2)
            self._is_read_battA = True

    def _read_flt_time(self):

        current_flight_time = time.time() - self._flight_time_start
        self._fltTime.data = int(current_flight_time)

        min = self._fltTime.data / 60
        sec = self._fltTime.data % 60

    def _read_gyro(self):

        self._is_read_gyro = False
        if (self._vehicle.attitude.roll is not None and
                self._vehicle.attitude.pitch is not None and
                self._vehicle.attitude.yaw is not None):
            self._gyro.tm = RTC.Time(0, 0)
            self._gyro.data = RTC.Orientation3D(self._vehicle.attitude.roll,
                                                self._vehicle.attitude.pitch,
                                                self._vehicle.attitude.yaw)
            self._is_read_gyro = True

    def _write_gyro(self):

        if self._is_read_gyro:
            self._d_outGyro.data = self._gyro.data
            self._outGyroOut.write()

    def _read_accel(self):

        self._is_read_accel = False
        accel = self._vehicle.velocity
        if all(elem is not None for elem in accel):
            self._accel.tm = RTC.Time(0, 0)
            self._accel.data = accel
            self._is_read_accel = True

    def _read_mag(self):

        self._is_read_mag = False
        if self._vehicle.heading is not None:
            self._mag.tm = RTC.Time(0, 0)
            self._mag.data = self._vehicle.heading
            self._is_read_mag = True

    def _read_temperature(self):

        self._is_read_temp = False
        temp = self._vehicle.scaled_pres.temperature
        if temp is not None:
            self._temp.tm = RTC.Time(0, 0)
            self._temp.data = round(temp, 2)
            self._is_read_temp = True

    def _read_pressure(self):

        self._is_read_pres = False
        pres = self._vehicle.scaled_pres.press_abs
        if pres is not None:
            self._pres.tm = RTC.Time(0, 0)
            self._pres.data = round(pres, 2)
            self._is_read_pres = True

    def _read_flt_mode(self):

        self._fltMode.tm = RTC.Time(0, 0)
        self._fltMode.data = self._vehicle.mode.name

    def _read_io_state(self):

        self._state.tm = RTC.Time(0, 0)
        self._state.data = droneio_sts_str[self._drone_sts]

    def _set_io_err_inf(self, msg):
    
        self._sysErr.tm = RTC.Time(0, 0)
        if msg == ERR_MSG_008:
            self._sysErr.data += msg
        else:
            self._sysErr.data = msg
        if len(msg) != 0:
            logger.error("%s", msg)

    def _write_io_err_inf(self, msg):

        self._sysErr.tm = RTC.Time(0, 0)
        self._sysErr.data = msg
        self._d_outDroneView.sysErr.data = self._sysErr.data
        if self._outDroneViewOut.write():
            if len(msg) != 0:
                logger.error("%s", msg)

    def _set_drone_ctrl(self):

        self._d_outDroneCtrl.id += 1
        self._d_outDroneCtrl.robotID.data = self._RobotID[0]
        self._d_outDroneCtrl.state.data = self._state.data

        if self._is_read_gps_home and self._is_read_gps_drone:
                self._d_outDroneCtrl.homeGPS.latitude = self._gps_home.latitude
                self._d_outDroneCtrl.homeGPS.longitude = self._gps_home.longitude
                self._d_outDroneCtrl.homeGPS.altitude = self._gps_home.altitude
                self._d_outDroneCtrl.droneGPS.latitude = self._gps_drone.latitude
                self._d_outDroneCtrl.droneGPS.longitude = self._gps_drone.longitude
                self._d_outDroneCtrl.droneGPS.altitude = self._gps_drone.altitude

        if self._is_read_mag:
            self._d_outDroneCtrl.mag.data = self._mag.data

        if self._is_read_accel:
            self._d_outDroneCtrl.accel.data = self._accel.data

    def _write_drone_ctrl(self):

        OpenRTM_aist.setTimestamp(self._d_outDroneCtrl)
        if self._outDroneCtrlOut.write():
            logger.debug("SendID:   %d", self._d_outDroneCtrl.id)
            logger.debug("Home:     %.10f %.10f %.10f",
                        self._d_outDroneCtrl.homeGPS.latitude,
                        self._d_outDroneCtrl.homeGPS.longitude,
                        self._d_outDroneCtrl.homeGPS.altitude)
            logger.debug("Drone:    %.10f %.10f %.10f",
                        self._d_outDroneCtrl.droneGPS.latitude,
                        self._d_outDroneCtrl.droneGPS.longitude,
                        self._d_outDroneCtrl.droneGPS.altitude)
            logger.debug("Accel:    %s", self._d_outDroneCtrl.accel.data)
            logger.debug("Mag:      %d", self._d_outDroneCtrl.mag.data)
            logger.debug("State:    %s", self._d_outDroneCtrl.state.data)

    def _set_map(self, rtc_state):

        if self._is_read_gps_drone:
            if self._is_read_mag:
                self._d_outDroneMap.tm = RTC.Time(0, 0)
                self._d_outDroneMap.droneGPS.latitude = self._gps_drone.latitude
                self._d_outDroneMap.droneGPS.longitude = self._gps_drone.longitude
                self._d_outDroneMap.droneGPS.altitude = self._gps_drone.altitude
                if rtc_state == "Error":
                    self._d_outDroneMap.targetGPS.latitude = self._gps_home.latitude
                    self._d_outDroneMap.targetGPS.longitude = self._gps_home.longitude
                    self._d_outDroneMap.targetGPS.altitude = self._gps_home.altitude
                elif rtc_state == "Active":
                    if self._d_inTarget.pos.latitude != 100.0 and self._d_inTarget.pos.longitude != 200.0 and self._d_inTarget.pos.altitude != 0:
                        self._d_outDroneMap.targetGPS.latitude = self._d_inTarget.pos.latitude
                        self._d_outDroneMap.targetGPS.longitude = self._d_inTarget.pos.longitude
                        self._d_outDroneMap.targetGPS.altitude = self._d_inTarget.pos.altitude
                self._d_outDroneMap.mag.data = self._mag.data
                self._d_outDroneMap.state.data = self._state.data

    def _write_map(self):

        if self._is_read_gps_drone:
            if self._is_read_mag:
                self._outDroneMapOut.write()

    def _set_view(self):
    
        if self._is_read_gps_home:
            self._d_outDroneView.homeGPS.latitude = self._gps_home.latitude
            self._d_outDroneView.homeGPS.longitude = self._gps_home.longitude
            self._d_outDroneView.homeGPS.altitude = self._gps_home.altitude

        if self._is_read_gps_drone:
            self._d_outDroneView.droneGPS.latitude = self._gps_drone.latitude
            self._d_outDroneView.droneGPS.longitude = self._gps_drone.longitude
            self._d_outDroneView.droneGPS.altitude = self._gps_drone.altitude
            self._d_outDroneView.targetSpeed.data = self._d_inTarget.speed.data

        if self._is_read_battV:
            self._d_outDroneView.batt.data[0] = self._batt.data[0]

        if self._is_read_battA:
            self._d_outDroneView.batt.data[1] = self._batt.data[1]

        self._d_outDroneView.fltTime.data = self._fltTime.data

        if self._is_read_gyro:
            self._d_outDroneView.gyro.data = self._gyro.data
 
        if self._is_read_accel:
            self._d_outDroneView.accel.data = self._accel.data

        if self._is_read_mag:
            self._d_outDroneView.mag.data = self._mag.data

        if self._is_read_temp:
            self._d_outDroneView.temp.data = self._temp.data

        if self._is_read_pres:
            self._d_outDroneView.pres.data = self._pres.data

        self._d_outDroneView.fltMode.data = self._fltMode.data
        self._d_outDroneView.state.data = self._state.data
        self._d_outDroneView.sysErr.data = self._sysErr.data

    def _write_view(self):
    
        if self._outDroneViewOut.write():
            logger.debug("===== viewer s =====")        
            logger.debug("Home   %.10f %.10f %.10f",
                        self._d_outDroneView.homeGPS.latitude,
                        self._d_outDroneView.homeGPS.longitude,
                        self._d_outDroneView.homeGPS.altitude)
            logger.debug("Drone  %.10f %.10f %.10f",
                        self._d_outDroneView.droneGPS.latitude,
                        self._d_outDroneView.droneGPS.longitude,
                        self._d_outDroneView.droneGPS.altitude)
            logger.debug("tgtSpd %.2f", self._d_outDroneView.targetSpeed.data)
            logger.debug("Batt   %.2f V %.2f A",
                        self._d_outDroneView.batt.data[0],
                        self._d_outDroneView.batt.data[1])
            logger.debug("Time   %d:%02d",
                        self._d_outDroneView.fltTime.data / 60,
                        self._d_outDroneView.fltTime.data % 60)
            logger.debug("Gyro   r:%f p:%f y:%f",
                        self._d_outDroneView.gyro.data.r,
                        self._d_outDroneView.gyro.data.p,
                        self._d_outDroneView.gyro.data.y)
            logger.debug("Accel  x:%.2f y:%.2f z:%.2f",
                        self._d_outDroneView.accel.data[0],
                        self._d_outDroneView.accel.data[1],
                        self._d_outDroneView.accel.data[2])
            logger.debug("Mag    %d", self._d_outDroneView.mag.data)
            logger.debug("Baro   temp:%.2f pres:%.2f", self._d_outDroneView.temp.data, self._d_outDroneView.pres.data)
            logger.debug("FltMd  %s", self._d_outDroneView.fltMode.data)
            logger.debug("State  %s", self._d_outDroneView.state.data)
            if len(self._d_outDroneView.sysErr.data) != 0:
                logger.debug("Err    %s", self._d_outDroneView.sysErr.data)
            logger.debug("===== viewer e =====")        

    def _set_db(self):
    
        self._d_outDroneDB.robotID.data = self._RobotID[0]

        if self._is_read_gps_home:
            self._d_outDroneDB.homeGPS.latitude = self._gps_home.latitude
            self._d_outDroneDB.homeGPS.longitude = self._gps_home.longitude
            self._d_outDroneDB.homeGPS.altitude = self._gps_home.altitude
        else:
            self._d_outDroneDB.homeGPS.latitude = 100.0
            self._d_outDroneDB.homeGPS.longitude = 200.0
            self._d_outDroneDB.homeGPS.altitude = 0.0

        if self._is_read_gps_drone:
            self._d_outDroneDB.droneGPS.latitude = self._gps_drone.latitude
            self._d_outDroneDB.droneGPS.longitude = self._gps_drone.longitude
            self._d_outDroneDB.droneGPS.altitude = self._gps_drone.altitude
        else:
            self._d_outDroneDB.droneGPS.latitude = 100.0
            self._d_outDroneDB.droneGPS.longitude = 200.0
            self._d_outDroneDB.droneGPS.altitude = 0.0

        if self._is_read_battV:
            self._d_outDroneDB.battV.data = self._batt.data[0]
        else:
            self._d_outDroneDB.battV.data = 0.0

        if self._is_read_battA:
            self._d_outDroneDB.battA.data = self._batt.data[1]
        else:
            self._d_outDroneDB.battA.data = 0.0

        self._d_outDroneDB.fltTime.data = self._fltTime.data

        if self._is_read_gyro:
            self._d_outDroneDB.gyro.data.r = self._gyro.data.r
            self._d_outDroneDB.gyro.data.p = self._gyro.data.p
            self._d_outDroneDB.gyro.data.y = self._gyro.data.y
            self._d_outDroneDB.correctIMU.data[3] = self._d_outDroneDB.gyro.data.r
            self._d_outDroneDB.correctIMU.data[4] = self._d_outDroneDB.gyro.data.p
            self._d_outDroneDB.correctIMU.data[5] = self._d_outDroneDB.gyro.data.y
        else:
            self._d_outDroneDB.gyro.data = RTC.Orientation3D(0.0, 0.0, 0.0)
            self._d_outDroneDB.correctIMU.data[3] = 0.0
            self._d_outDroneDB.correctIMU.data[4] = 0.0
            self._d_outDroneDB.correctIMU.data[5] = 0.0

        if self._is_read_accel:
            self._d_outDroneDB.accel.data[0] = self._accel.data[0]
            self._d_outDroneDB.accel.data[1] = self._accel.data[1]
            self._d_outDroneDB.accel.data[2] = self._accel.data[2]
            self._d_outDroneDB.correctIMU.data[0] = self._d_outDroneDB.accel.data[0]
            self._d_outDroneDB.correctIMU.data[1] = self._d_outDroneDB.accel.data[1]
            self._d_outDroneDB.correctIMU.data[2] = self._d_outDroneDB.accel.data[2]
        else:
            self._d_outDroneDB.accel.data = [0.0, 0.0, 0.0]
            self._d_outDroneDB.correctIMU.data[0] = 0.0
            self._d_outDroneDB.correctIMU.data[1] = 0.0
            self._d_outDroneDB.correctIMU.data[2] = 0.0

        if self._is_read_mag:
            self._d_outDroneDB.mag.data = self._mag.data
            self._d_outDroneDB.correctIMU.data[6] = self._d_outDroneDB.mag.data
        else:
            self._d_outDroneDB.mag.data = 0.0
            self._d_outDroneDB.correctIMU.data[6] = 0.0

        if self._is_read_temp:
            self._d_outDroneDB.temp.data = self._temp.data
        else:
            self._d_outDroneDB.temp.data = 0.0

        if self._is_read_pres:
            self._d_outDroneDB.pres.data = self._pres.data
        else:
            self._d_outDroneDB.pres.data = 0.0

        self._d_outDroneDB.fltMode.data = self._fltMode.data
        self._d_outDroneDB.state.data = self._state.data
        self._d_outDroneDB.sysErr.data = self._sysErr.data

        self._d_outDroneDB.rawIMU.data = [
            self._vehicle.raw_imu.xacc,
            self._vehicle.raw_imu.yacc,
            self._vehicle.raw_imu.zacc,
            self._vehicle.raw_imu.xgyro,
            self._vehicle.raw_imu.ygyro,
            self._vehicle.raw_imu.zgyro,
            self._vehicle.raw_imu.xmag,
            self._vehicle.raw_imu.ymag,
            self._vehicle.raw_imu.zmag]

    def _write_db(self):
    
        # sender time
        OpenRTM_aist.setTimestamp(self._d_outDroneDB)
        self._d_outDroneDB.robotID.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.homeGPS.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.droneGPS.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.battV.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.battA.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.fltTime.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.gyro.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.accel.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.mag.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.temp.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.pres.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.fltMode.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.state.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.sysErr.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.rawIMU.tm = self._d_outDroneDB.tm
        self._d_outDroneDB.correctIMU.tm = self._d_outDroneDB.tm

        if self._outDroneDBOut.write():
            logger.debug("%s", self._d_outDroneDB)

    def _write_io_state(self):

        self._state.data = droneio_sts_str[self._drone_sts]
        
        # DroneCtrl
        self._d_outDroneCtrl.id += 1
        self._d_outDroneCtrl.robotID.data = self._RobotID[0]
        self._d_outDroneCtrl.state.data = self._state.data

        # DroneMap
        self._d_outDroneMap.tm = RTC.Time(0, 0)
        self._d_outDroneMap.state.data = self._state.data

        # DroneView
        self._d_outDroneView.tm = RTC.Time(0, 0)
        self._d_outDroneView.state.data = self._state.data

        OpenRTM_aist.setTimestamp(self._d_outDroneCtrl)

        ctrl_out_write = False
        if self._outDroneCtrlOut.write():
            ctrl_out_write = True
        self._outDroneMapOut.write()
        self._outDroneViewOut.write()

        if ctrl_out_write:
            logger.debug("SendID:   %d", self._d_outDroneCtrl.id)
            logger.debug("State:    %s", self._d_outDroneCtrl.state.data)

    def _is_batt_state(self):

        rtn = True

        if self._BattChk[0] == 'on':
            if(self._vehicle.battery.voltage < self._batt_thr):
                rtn = False

        return rtn

    def _is_valid_lat(self, lat):

        rtn = False

        if -90.0 <= lat <= 90.0:
            rtn = True
        else:
            logger.error("[E]Invalid lat(%f[deg]) received!!", lat)

        return rtn

    def _is_valid_lon(self, lon):

        rtn = False

        if -180.0 <= lon <= 180.0:
            rtn = True
        else:
            logger.error("[E]Invalid lon(%f[deg]) received!!", lon)

        return rtn

    def _is_valid_alt(self, alt):

        rtn = False

        if self._drone_sts != STS_ARMED:
            if alt > 0:
                rtn = True
            else:
                logger.error("[E]Invalid alt(%f[m]) received!!", alt)
        else:
            rtn = True

        return rtn

    def _is_valid_yaw_deg(self, yaw):

        rtn = False

        if 0.0 <= yaw <= 360.0:
            rtn = True
        else:
            logger.error("[E]Invalid yaw(%f[deg]) received!!", yaw)

        return rtn

    def _ready_to_log_write(self, file_mode):
        
        if file_mode == FILE_MODE_WRITE:
            date, time = self._get_datetime_to_str()
            time = time.replace(":", "").replace(".", "")  #  HH:mm:ss.sss --> hhmmssSSS
            self._file_name = date + "_" + time[0:6] + "_%s_log.csv" % (self._RobotID[0])

        if self._is_open_rtc_viewer_log(file_mode):
            if self._is_csv_writer():
                if file_mode == FILE_MODE_WRITE:
                    self._writer.writerow(self._header)
                    self._write_rtc_viewer_log("Active")
        
    def _get_datetime_to_str(self):
        
        curr_time = datetime.now()
        date = "%s%s%s" % ("{0:04d}".format(curr_time.year), "{0:02d}".format(curr_time.month), "{0:02d}".format(curr_time.day))
        time = "%s:%s:%s.%s" % ("{0:02d}".format(curr_time.hour), "{0:02d}".format(curr_time.minute),
                           "{0:02d}".format(curr_time.second), "{0:03d}".format(curr_time.microsecond / 1000))
        return date, time

    def _is_open_rtc_viewer_log(self, file_mode):

        rtn = False
        
        try:
            if self._file_name != "":
                self._file = open(self._file_name, file_mode)
                rtn = True
        except IOError as e:
            logger.error("I/O Error {0} {1}".format(e.errno, e.strerror))

        return rtn
                
    def _is_csv_writer(self):

        rtn = False

        try:
            if self._file is not None:
                self._writer = csv.writer(self._file)
                if self._writer is not None:
                    rtn = True
        except csv.Error as e:
            logger.error("CSV writer error {0} {1}".format(e.errno, e.strerror))

        return rtn

    def _write_rtc_viewer_log(self, rtc_state):

        if self._writer is not None:
            self._set_log(rtc_state)
            self._writer.writerow(self._log_row)
    
    def _close_rtc_viewer_log(self):

        if self._file is not None:
            self._file.close()
            self._file = None
            self._writer = None

    def _conv_rtctime_to_datetime(self, tm):

        curr_time = datetime.fromtimestamp(tm.sec)

        date = "%s%s%s" % ("{0:04d}".format(curr_time.year), "{0:02d}".format(curr_time.month),
                           "{0:02d}".format(curr_time.day))

        time = "%s:%s:%s.%s" % ("{0:02d}".format(curr_time.hour), "{0:02d}".format(curr_time.minute),
                                "{0:02d}".format(curr_time.second), "{0:03d}".format(tm.nsec / 1000000))

        return date, time

    def _set_log(self, rtc_state):
    
        self._set_rcv_log(rtc_state)
        self._set_snd_log(rtc_state)
        self._set_sensor_log(rtc_state)
        self._set_other_log(rtc_state)

    def _set_rcv_log(self, rtc_state):

        if rtc_state == "Active":
            if self._d_inTarget.id != 0:
                rcv_date, rcv_time = self._conv_rtctime_to_datetime(self._d_inTarget.tm)

                self._log_row[self._header.index("RcvDate")] = rcv_date
                self._log_row[self._header.index("RcvTime")] = rcv_time
                self._log_row[self._header.index("RcvRobotID")] = "GCS"
                self._log_row[self._header.index("RcvID")] = self._d_inTarget.id
                self._log_row[self._header.index("TargetLat")] = "{0:.10f}".format(self._d_inTarget.pos.latitude)
                self._log_row[self._header.index("TargetLon")] = "{0:.10f}".format(self._d_inTarget.pos.longitude)
                self._log_row[self._header.index("TargetAlt")] = "{0:.2f}".format(self._d_inTarget.pos.altitude)
                self._log_row[self._header.index("TargetSpd")] = "{0:.2f}".format(self._d_inTarget.speed.data)
                self._log_row[self._header.index("TargetYawDir")] = self._d_inTarget.yaw.data[0]
                self._log_row[self._header.index("TargetYawDeg")] = self._d_inTarget.yaw.data[1]

    def _set_snd_log(self, rtc_state):

        if rtc_state != "Exit":
            snd_date, snd_time = self._conv_rtctime_to_datetime(self._d_outDroneCtrl.tm)

            self._log_row[self._header.index("SndDate")] = snd_date
            self._log_row[self._header.index("SndTime")] = snd_time
            self._log_row[self._header.index("SndRobotID")] = "GCS"
            self._log_row[self._header.index("SndID")] = self._d_outDroneCtrl.id
            self._log_row[self._header.index("State")] = self._d_outDroneCtrl.state.data

            if self._is_read_gps_home and self._is_read_gps_drone:
                self._log_row[self._header.index("HomeLat")] = "{0:.10f}".format(self._d_outDroneCtrl.homeGPS.latitude)
                self._log_row[self._header.index("HomeLon")] = "{0:.10f}".format(self._d_outDroneCtrl.homeGPS.longitude)
                self._log_row[self._header.index("HomeAlt")] = "{0:.2f}".format(self._d_outDroneCtrl.homeGPS.altitude)
                self._log_row[self._header.index("RobotLat")] = "{0:.10f}".format(self._d_outDroneCtrl.droneGPS.latitude)
                self._log_row[self._header.index("RobotLon")] = "{0:.10f}".format(self._d_outDroneCtrl.droneGPS.longitude)
                self._log_row[self._header.index("RobotAlt")] = "{0:.2f}".format(self._d_outDroneCtrl.droneGPS.altitude)

            if self._is_read_accel:
                self._log_row[self._header.index("AccX")] = "{0:.2f}".format(self._d_outDroneCtrl.accel.data[0])
                self._log_row[self._header.index("AccY")] = "{0:.2f}".format(self._d_outDroneCtrl.accel.data[1])
                self._log_row[self._header.index("AccZ")] = "{0:.2f}".format(self._d_outDroneCtrl.accel.data[2])

            if self._is_read_mag:
                self._log_row[self._header.index("Mag")] = int(self._d_outDroneCtrl.mag.data)

    def _set_sensor_log(self, rtc_state):

        if rtc_state in ["Active", "Error"]:
            if self._is_read_battV:
                self._log_row[self._header.index("BattV")] = "{0:.2f}".format(self._batt.data[0])

            if self._is_read_gyro:
                self._log_row[self._header.index("GyroR")] = "{0:.3f}".format(self._gyro.data.r)
                self._log_row[self._header.index("GyroP")] = "{0:.3f}".format(self._gyro.data.p)
                self._log_row[self._header.index("GyroY")] = "{0:.3f}".format(self._gyro.data.y)

            if self._is_read_temp:
                self._log_row[self._header.index("Temp")] = "{0:.2f}".format(self._temp.data)

            if self._is_read_pres:
                self._log_row[self._header.index("Pres")] = "{0:.2f}".format(self._pres.data)

    def _set_other_log(self, rtc_state):
  
        self._log_row[self._header.index("Category")] = self._RobotID[0]
        self._log_row[self._header.index("ActStartDate")] = self._act_start_date
        self._log_row[self._header.index("ActStartTime")] = self._act_start_time
        self._log_row[self._header.index("FltMode")] = self._fltMode.data
        self._log_row[self._header.index("RTCState")] = rtc_state
        if len(self._sysErr.data) != 0:
            self._log_row[self._header.index("Err")] = self._sysErr.data


class RawIMU(object):

    def __init__(self, time_boot_us=0, xacc=0, yacc=0, zacc=0, xgyro=0, ygyro=0, zgyro=0, xmag=0, ymag=0, zmag=0):
        self._time_boot_us = time_boot_us
        self._xacc = xacc
        self._yacc = yacc
        self._zacc = zacc
        self._xgyro = xgyro
        self._ygyro = ygyro
        self._zgyro = zgyro
        self._xmag = xmag
        self._ymag = ymag
        self._zmag = zmag

    def __str__(self):
        return "xacc={},yacc={},zacc={},xgyro={},ygyro={},zgyro={},xmag={},ymag={},zmag={}".format(self._xacc, self._yacc, self._zacc, self._xgyro, self._ygyro, self._zgyro, self._xmag, self._ymag, self._zmag)


class Scaled_Pressure(object):

    def __init__(self, time_boot_ms=0.0, press_abs=0.0, press_diff=0.0, temperature=0.0):
        self._time_boot_ms = time_boot_ms
        self._press_abs = press_abs
        self._press_diff = press_diff
        self._temperature = temperature

    def __str__(self):
        return "press_abs={},temperature={}".format(self._press_abs, self._temperature)


class MyVehicle(Vehicle):

    def __init__(self, *args):
        super(MyVehicle, self).__init__(*args)

        # Create an Vehicle.raw_imu object with initial values set to None.
        self._raw_imu = RawIMU()

        # Create a message listener using the decorator.
        @self.on_message('RAW_IMU')
        def listener(self, name, message):
            self._raw_imu.time_boot_us = message.time_usec
            self._raw_imu.xacc = message.xacc
            self._raw_imu.yacc = message.yacc
            self._raw_imu.zacc = message.zacc
            self._raw_imu.xgyro = message.xgyro
            self._raw_imu.ygyro = message.ygyro
            self._raw_imu.zgyro = message.zgyro
            self._raw_imu.xmag = message.xmag
            self._raw_imu.ymag = message.ymag
            self._raw_imu.zmag = message.zmag

            self.notify_attribute_listeners('raw_imu', self._raw_imu)

        # Create an Vehicle.'scaled_pres' object with initial values set to
        # None.
        self._scaled_pres = Scaled_Pressure()

        # Create a message listener using the decorator.
        @self.on_message('SCALED_PRESSURE')
        def listener(self, name, message):
            self._scaled_pres.time_boot_ms = message.time_boot_ms
            self._scaled_pres.press_abs = message.press_abs
            self._scaled_pres.press_diff = message.press_diff
            self._scaled_pres.temperature = message.temperature * 0.01

            self.notify_attribute_listeners('scaled_pres', self._scaled_pres)

    @property
    def raw_imu(self):
        return self._raw_imu

    @property
    def scaled_pres(self):
        return self._scaled_pres


def DroneIOInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=droneio_spec)
    manager.registerFactory(profile,
                            DroneIO,
                            OpenRTM_aist.Delete)


def MyModuleInit(manager):
    DroneIOInit(manager)

    # Create a component
    comp = manager.createComponent("DroneIO")


def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager()

if __name__ == "__main__":
    main()
