#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file RobotRPiCam.py
 @brief Periodically generate and save images and additional data
 @date $Date$

 @author East Japan Accounting Center Co., Ltd.

 (c)2017 The University of Aizu
 This software is released under the New BSD License.

"""
import sys
import time
import datetime
import os
import math
from decimal import *
import ConfigParser
import subprocess
import commands
from picamera import PiCamera
sys.path.append(".")

# Import RTM module
import RTC
import OpenRTM_aist

# logger
from logging import getLogger, StreamHandler, DEBUG
logger = getLogger(__name__)
handler = StreamHandler()
handler.setLevel(DEBUG)
logger.setLevel(DEBUG)
logger.addHandler(handler)

# ini-file name
INI_FILE = "sno.ini"

# ini-file section
INI_SECT = "last"

# ini-file date
INI_DAY = "day"

# ini-file section num
INI_SNO = "sno"

# image save directory
IMAGE_FOLDER = "Picture/"

# image-additional save directory
ADD_DATA_FOLDER = "Additional data/"

# image additional separation
SEP = ","

# image extension
IMAGE_EXT = ".jpg"

# image-additional extension
ADD_DATA_EXT = ".csv"

# max length of shooting plan name
MAX_LEN = 30

# Error Message
ERR_MSG_01 = "ERROR: RPi camera is NOT available"
ERR_MSG_02 = "ERROR: Creation of save destination folder failed ( %(path)s )"
ERR_MSG_03 = "WARNING: Failed to save image"
ERR_MSG_04 = "WARNING: Failed to save csv data"

# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
robotrpicam_spec = ["implementation_id", "RobotRPiCam",
                    "type_name",         "RobotRPiCam",
                    "description",       "Periodically generate and save images and additional data",
                    "version",           "1.0.0",
                    "vendor",            "EAC",
                    "category",          "Camera",
                    "activity_type",     "STATIC",
                    "max_instance",      "1",
                    "language",          "Python",
                    "lang_type",         "SCRIPT",
                    "conf.default.Interval", "1.0",
                    "conf.default.Width", "3200",
                    "conf.default.Height", "2400",
                    "conf.default.RobotID", "drone1",
                    "conf.default.CameraNo", "1",
                    "conf.default.RootPath", "/home/pi/",
                    "conf.default.PlanName", "TestShot",
                    "conf.default.Rotation", "0",
                    "conf.__widget__.Interval", "spin",
                    "conf.__widget__.Width", "spin",
                    "conf.__widget__.Height", "spin",
                    "conf.__widget__.RobotID", "text",
                    "conf.__widget__.CameraNo", "spin",
                    "conf.__widget__.RootPath", "text",
                    "conf.__widget__.PlanName", "text",
                    "conf.__widget__.Rotation", "radio",
                    "conf.__constraints__.Interval", "x>=0.1",
                    "conf.__constraints__.Width", "x>=1",
                    "conf.__constraints__.Height", "x>=1",
                    "conf.__constraints__.Rotation", "(0,90,180,270)",
                    ""]
# </rtc-template>

##
# @class RobotRPiCam
# @brief Periodically generate and save images and additional data
#
# Overview:  Refer to the "機能仕様書_RobotRPiCam" for details.
#
# In / Out:  Refer to the "機能仕様書_RobotRPiCam" for details.
#
# Algorythm:  Refer to the "機能仕様書_RobotRPiCam" for details.
#
#


class RobotRPiCam(OpenRTM_aist.DataFlowComponentBase):

    ##
    # @brief constructor
    # @param manager Maneger Object
    #
    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_inGPSRobot = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(
            0, 0), 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        Robot GPS
         - Type: RTC::GPSData
         - Semantics: - Latitude
                      - Longitude
                      - Altitude
         - Unit: deg
        """
        self._inGPSRobotIn = OpenRTM_aist.InPort("inGPSRobot", self._d_inGPSRobot)
        self._d_inGyro = RTC.TimedOrientation3D(RTC.Time(0, 0), 0)
        """
        Robot gyro
         - Type: RTC::TimedOrientation3D
         - Semantics: - Roll Axis
                      - Pitch Axis
                      - Yaw Axis
        """
        self._inGyroIn = OpenRTM_aist.InPort("inGyro", self._d_inGyro)

        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">
        """
        image capture interval
         - Name: Interval Interval
         - DefaultValue: 1.0
         - Unit: sec
         - Range: 0.1<=x
        """
        self._Interval = [1.0]
        """
        image width
         - Name: Width Width
         - DefaultValue: 3200
         - Unit: px
         - Range: 1<=x
        """
        self._Width = [3200]
        """
        image height
         - Name: Height Height
         - DefaultValue: 2400
         - Unit: px
         - Range: 1<=x
        """
        self._Height = [2400]
        """
        robot id
         - Name: RobotID RobotID
         - DefaultValue: drone1
        """
        self._RobotID = ['drone1']
        """
        camera id
         - Name: CameraID CameraNo
         - DefaultValue: 1
        """
        self._CameraNo = [1]
        """
        Root path of image and additional data
         - Name: RootPath RootPath
         - DefaultValue: /home/pi/
        """
        self._RootPath = ['/home/pi/']
        """
        Shooting plan name (30 half-width characters)
         - Name: string PlanName
         - DefaultValue: TestShot
        """
        self._PlanName = ['TestShot']
        """
        Set the camera image rotation.
        Valid values are 0, 90, 180, 270 only.
         - Name: short Rotation
         - DefaultValue: 0
        """
        self._Rotation = [0]

        # </rtc-template>

    ##
    # RTC initialization
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry()
    #
    # @return RTC::ReturnCode_t
    #
    #
    def onInitialize(self):
        # Bind variables and configuration variable
        self.bindParameter("Interval", self._Interval, "1.0")
        self.bindParameter("Width", self._Width, "3200")
        self.bindParameter("Height", self._Height, "2400")
        self.bindParameter("RobotID", self._RobotID, "drone1")
        self.bindParameter("CameraNo", self._CameraNo, "1")
        self.bindParameter("RootPath", self._RootPath, "/home/pi/")
        self.bindParameter("PlanName", self._PlanName, "TestShot")
        self.bindParameter("Rotation", self._Rotation, "0")

        # Set InPort buffers
        self.addInPort("inGPSRobot", self._inGPSRobotIn)
        self.addInPort("inGyro", self._inGyroIn)

        # Set OutPort buffers

        # Set service provider to Ports

        # Set service consumers to Ports

        # Set CORBA Service Ports

        self._camera = PiCamera()

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The finalize action (on ALIVE->END transition)
    #    # formaer rtc_exiting_entry()
    #    #
    #    # @return RTC::ReturnCode_t
    #
    #    #
    # def onFinalize(self):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The startup action when ExecutionContext startup
    #    # former rtc_starting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onStartup(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The shutdown action when ExecutionContext stop
    #    # former rtc_stopping_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onShutdown(self, ec_id):
    #
    #    return RTC.RTC_OK

        ##
        # Raspberry pi Camera connection check and directory creation.
        #
        # The activated action (Active state entry action)
        # former rtc_active_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onActivated(self, ec_id):

        res = True

        self.robotID = self._RobotID[0]
        self.rootPath = self._RootPath[0]
        self.camNo = self._CameraNo[0]
        self.planName = self.str_len_check(self._PlanName[0])

        # get the datetime
        self.today = datetime.date.today()

        # get the section no.
        self.sectionNo = self.get_section_no()

        # get the now time
        self.shotTime = self.get_now_time()

        # create the image dir
        imagePath = self.create_image_path()
        try:
            os.makedirs(imagePath)
        except:
            res = False
            logger.error(ERR_MSG_02 % {"path": imagePath})

        if res:
            # create the image additional dir
            adddataPath = self.create_adddata_path()
            try:
                os.makedirs(adddataPath)
            except:
                res = False
                logger.error(ERR_MSG_02 % {"path": adddataPath})

        if res:
            # update the ini-file section no
            self.set_section_no()

            # Raspi-camera connect check
            getcaminf = commands.getoutput("vcgencmd get_camera").split()
            raspicam_support = getcaminf[0].split('=')
            raspicam_detect = getcaminf[1].split('=')
            if raspicam_support[1] == '0' or raspicam_detect[1] == '0':
                res = False
                logger.error(ERR_MSG_01)                        

        if res:
            self._camera.resolution = (self._Width[0], self._Height[0])
            self._camera.rotation = self._Rotation[0]
            self._camera.start_preview()
            return RTC.RTC_OK
        else:
            return RTC.RTC_ERROR


        ##
        # Raspberry Pi Camera preview stop
        #
        # The deactivated action (Active state exit action)
        # former rtc_active_exit()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onDeactivated(self, ec_id):

        self._camera.stop_preview()

        return RTC.RTC_OK

        ##
        # Generate and save images and image additional data taken with Raspberry pi
    #    camera
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onExecute(self, ec_id):

        # get the GPS-info
        res = self.read_gps_robot()

        # Start the capture, if GPS is valid
        if res:
            nowTime = self.get_now_time()
            timer = float(nowTime) - float(self.shotTime)
            if timer >= self._Interval[0]:

                # get the gyro-info
                self.read_gyro()

                # Conv the time-info
                self.setTime = datetime.datetime.fromtimestamp(
                    float(nowTime))

                # setting the image filename
                imageName = self.create_file_name() + IMAGE_EXT

                # setting the image-additional filename
                addDataName = self.create_file_name() + ADD_DATA_EXT

                # Create the image-additional
                addData = self.create_adddata()

                # Image capture
                try:
                    self._camera.capture("%s" % (self.create_image_path() + imageName))
                except:
                    logger.warning(ERR_MSG_03)
                    res = False

                # Image-additional data is not saved, if image capture is failed
                if res:
                    # Save the image-additional
                    try:
                        with open(self.create_adddata_path() + addDataName, "w") as createText:
                            createText.write(addData)
                    except:
                        logger.warning(ERR_MSG_04)

                self.shotTime = nowTime

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The aborting action when main logic error occurred.
    #    # former rtc_aborting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onAborting(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The error action in ERROR state
    #    # former rtc_error_do()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onError(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The reset action that is invoked resetting
    #    # This is same but different the former rtc_init_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onReset(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The state update action that is invoked after onExecute() action
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #

    #    #
    # def onStateUpdate(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The action that is invoked when execution context's rate is changed
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onRateChanged(self, ec_id):
    #
    #    return RTC.RTC_OK

    def get_now_time(self):
        return "%0.3f" % time.time()

    def get_section_no(self):
        # Read the ini-file
        inifile = ConfigParser.SafeConfigParser()
        inifile.read(INI_FILE)

        # Get the last image capture date from ini-file
        lastDay = inifile.get(INI_SECT, INI_DAY)

        # Update the section no, if today and last image capture date is different
        if lastDay != str(self.today):
            inifile.set(INI_SECT, INI_SNO, str(1))
            inifile.set(INI_SECT, INI_DAY, str(self.today))
            with open("./" + INI_FILE, "w") as createIni:
                inifile.write(createIni)
            inifile.read(INI_FILE)

        return inifile.getint(INI_SECT, INI_SNO)

    def set_section_no(self):
        inifile = ConfigParser.SafeConfigParser()
        inifile.read(INI_FILE)
        inifile.set(INI_SECT, INI_SNO, str(self.sectionNo + 1))
        with open("./" + INI_FILE, "w") as createIni:
            inifile.write(createIni)

    def read_gps_robot(self):
        if self._inGPSRobotIn.isNew():
            gpsRobot = self._inGPSRobotIn.read()
            self.lat = gpsRobot.latitude
            self.lon = gpsRobot.longitude
            self.alt = gpsRobot.altitude
            ret = True
        else:
            ret = False

        return ret

    def read_gyro(self):
        if self._inGyroIn.isNew():
            gyro = self._inGyroIn.read()
            self.roll = gyro.data.r
            self.pitch = gyro.data.p
            self.yaw = gyro.data.y
        else:
            self.roll = 0.0
            self.pitch = 0.0
            self.yaw = 0.0

    def create_folder_path(self):
        return self.robotID + self.today.strftime("_%Y%m%d_") + str(self.sectionNo)

    def create_image_path(self):
        return self.rootPath + IMAGE_FOLDER + self.create_folder_path() + "/"

    def create_adddata_path(self):
        return self.rootPath + ADD_DATA_FOLDER + self.create_folder_path() + "/"

    def create_file_name(self):
        ret = self.today.strftime("%Y%m%d_") + \
            self.setTime.strftime("%H%M%S") + "%03d" % (self.setTime.microsecond // 1000) + "_" + \
            self.robotID + "_" + str(self.camNo)

        return ret

    def create_adddata(self):
        ret = self.create_file_name() + IMAGE_EXT + SEP + \
            self.create_folder_path() + SEP + \
            str(self._Width[0]) + SEP + \
            str(self._Height[0]) + SEP + \
            self.robotID + SEP + \
            str(self.camNo) + SEP + \
            str(self.lat_range_check(self.ceiling_digit(self.lat))) + SEP + \
            str(self.range_check(self.ceiling_digit(self.lon))) + SEP + \
            str(self.ceiling_digit(self.alt)) + SEP + \
            str(self.rad_to_deg_yaw(self.yaw)) + SEP + \
            str(self.range_check(math.degrees(self.pitch))) + SEP + \
            str(self.range_check(math.degrees(self.roll))) + SEP + \
            self.today.strftime("%Y%m%d") + SEP + \
            self.setTime.strftime("%H%M%S") + "%03d" % (self.setTime.microsecond // 1000) + SEP + \
            str(self.planName)

        return ret

    # Round up to 6 decimal places
    def ceiling_digit(self, data):
        ret = Decimal(data).quantize(
            Decimal(".000000"), rounding=ROUND_CEILING)
        return ret

    def lat_range_check(self, lat):
        if lat > 90.0:
            lat = 90.0
        elif lat < -90.0:
            lat = -90.0

        return lat

    def range_check(self, range):
        if range > 180.0:
            ranRem = range % 360.0
            if ranRem > 180.0:
                range = ranRem - 360.0
            else:
                range = ranRem
        elif range < -180.0:
            ranRem = abs(range) % 360.0
            if ranRem > 180.0:
                range = 360.0 - ranRem
            else:
                range = -ranRem

        return range

    def rad_to_deg_yaw(self, rad):
        ret = math.degrees(rad)
        if ret > 360:
            ret = ret % 360
        elif ret < 0:
            ret = 360 - (abs(ret) % 360)

        return ret

    # Make unicode characters less than MAX_LEN characters
    def str_len_check(self, str):
        str = str.decode("UTF-8").encode("shift-jis")
        if len(str) > MAX_LEN:
            str = str[:MAX_LEN]
        str = str.decode("shift-jis").encode("UTF-8")

        return str



def RobotRPiCamInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=robotrpicam_spec)
    manager.registerFactory(profile,
                            RobotRPiCam,
                            OpenRTM_aist.Delete)


def MyModuleInit(manager):
    RobotRPiCamInit(manager)

    # Create a component
    comp = manager.createComponent("RobotRPiCam")


def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager()

if __name__ == "__main__":
    main()
